# -*- coding: utf-8 -*-
'''
Feature, FeatureGenerator, FeatureArray classes
===============================================

`FeatureArray` is The main datatype used to communicate `state`s, `action`s,
and `reward`s, between objects in `reil`. `FeatureArray` is basically a
dictionary that contains instances of `Feature`.
`FeatureGenerator` allows for generating new `Feature` instances. It can
`generate` a new value or turn an input into a `Feature`. It enforces
`categorical` and `numerical` constraints, and produces `normalized` value.
'''
from __future__ import annotations

import dataclasses
import itertools
from typing import (Any, Callable, Dict, Generic, Iterable, Literal, Optional,
                    Tuple, TypeVar, Union)

MISSING = '__missing_feature__'

T = TypeVar('T')
MissingType = Literal['__missing_feature__']


@dataclasses.dataclass(frozen=True)
class Feature(Generic[T]):
    '''
    Attributes
    ----------
    name:
        Name of the data.

    value:
        Value of the data. Can be one item, or a tuple of items of the same
        type.

    is_numerical:
        Is the value numerical?

    normalized:
        The normal form of the value.

    categories:
        A tuple of categories that the value can take.

    lower:
        The lower limit for numerical values.

    upper:
        The upper limit for numerical values.
    '''
    name: str
    value: Optional[Union[T, Tuple[T, ...], MissingType]] = None
    is_numerical: Optional[bool] = dataclasses.field(
        default=None, repr=False, compare=False)
    categories: Optional[Tuple[T, ...]] = dataclasses.field(
        default=None, repr=False, compare=False)
    lower: Optional[T] = dataclasses.field(
        default=None, repr=False, compare=False)
    upper: Optional[T] = dataclasses.field(
        default=None, repr=False, compare=False)
    normalized: Optional[
        Union[Tuple[float, ...], Tuple[int, ...]]] = dataclasses.field(
            default=None, repr=False, compare=False)
    dict_fields: Tuple[str, ...] = dataclasses.field(
        default=('name', 'value'), init=False, repr=False, compare=False)

    def __post_init__(self):
        if self.is_numerical is None:
            return

        if self.is_numerical:
            if self.categories is not None:
                raise ValueError('Numerical type cannot have categories.')

            self.__dict__['dict_fields'] = ('name', 'value', 'lower', 'upper')
        else:
            if self.lower is not None or self.upper is not None:
                raise ValueError(
                    'Categorical type cannot have lower and upper.')

            self.__dict__['dict_fields'] = ('name', 'value', 'categories')

    @classmethod
    def numerical(
            cls, name: str, value: Optional[Union[T, Tuple[T, ...]]] = None,
            lower: Optional[T] = None, upper: Optional[T] = None,
            normalized: Optional[Tuple[float, ...]] = None):
        '''Create a numerical instance of `Feature`.'''
        return cls(name=name, value=value, is_numerical=True,
                   lower=lower, upper=upper, normalized=normalized)

    @classmethod
    def categorical(
            cls, name: str,
            value: Optional[Union[T, Tuple[T, ...], MissingType]] = None,
            categories: Optional[Tuple[T, ...]] = None,
            normalized: Optional[Tuple[float, ...]] = None):
        '''Create a categorical instance of `Feature`.'''
        return cls(name=name, value=value, is_numerical=False,
                   categories=categories, normalized=normalized)

    def as_dict(self):
        '''
        Return the data as a dictionary.

        Returns
        -------
        :
            The data as a dictionary.
        '''
        return {field: self.__dict__[field] for field in self.dict_fields}

    def __add__(self, other: Any):
        my_type = type(self)
        if type(other) != my_type:
            raise TypeError("unsupported operand type(s) for +: "
                            f"'{my_type}' and '{type(other)}'")

        for k, v in self.__dict__.items():
            if k not in ('value', 'normalized'):
                if other.__dict__[k] != v:
                    raise TypeError(
                        f'Different {k} values: {v} != {other.__dict__[k]}.')

        new_value = self.__dict__['value'] + other.value
        if self.is_numerical:
            return my_type.numerical(
                name=self.name, value=new_value,
                lower=self.__dict__.get('lower'),
                upper=self.__dict__.get('upper'))
        else:
            return my_type.categorical(
                name=self.name, value=new_value,
                categories=self.__dict__.get('categories'))


@dataclasses.dataclass(frozen=True)
class FeatureGenerator(Generic[T]):
    '''
    A class to generate `Feature`s.

    Attributes
    ----------
    name:
        Name of the data.

    is_numerical:
        Is the feature to be generated numerical?

    categories:
        A tuple of categories that the value can take.

    probabilities:
        A tuple of probabilities corresponding with each category. This can
        be used to generate new random `Feature` instances.

    mean:
        The mean for numerical values. This can
        be used to generate new random `Feature` instances.

    stdev:
        The standard deviation for numerical values. This can
        be used to generate new random `Feature` instances.

    lower:
        The lower limit for numerical values.

    upper:
        The upper limit for numerical values.

    normalizer:
        For a categorical `FeatureGenerator`, normalizer is a dictionary of
        categories as keys and their corresponding one-hot encodings as values.
        For a numerical `FeatureGenerator`, normalizer is a function that
        accepts the value and returns its normalized value.

    randomized:
        Determines for the `generator`, whether the new `Feature` should be
        randomly generated.

    generator:
        A function that accepts a `FeatureGenerator` instance, and produces
        a value for the new `Feature`.

    allow_missing:
        If `True`, a categorical generator can generate a `MISSING` instance.
        Also the normalized form will have one more categories to account for
        `MISSING`.
    '''
    name: str
    is_numerical: bool = dataclasses.field(
        default=False, repr=False, compare=False)
    categories: Optional[Tuple[T, ...]] = None
    probabilities: Optional[Tuple[float, ...]] = None
    mean: Optional[T] = None
    stdev: Optional[T] = None
    lower: Optional[T] = None
    upper: Optional[T] = None
    normalizer: Optional[Any] = dataclasses.field(
        default=None, init=False, repr=False, compare=False)
    randomized: Optional[bool] = True
    generator: Optional[
        Callable[[FeatureGenerator[T]], Union[T, Tuple[T, ...]]]] = None
    allow_missing: bool = False

    @classmethod
    def numerical(
            cls, name: str,
            mean: Optional[float] = None, stdev: Optional[float] = None,
            lower: Optional[float] = None, upper: Optional[float] = None,
            generator: Optional[Callable[
                [FeatureGenerator[T]], Union[T, Tuple[T, ...]]]] = None,
            randomized: Optional[bool] = None):
        return cls(
            name=name, is_numerical=True, lower=lower, upper=upper, mean=mean,
            stdev=stdev, generator=generator, randomized=randomized,
            allow_missing=False)

    @classmethod
    def categorical(
            cls, name: str,
            categories: Optional[Tuple[T, ...]] = None,
            probabilities: Optional[Tuple[float, ...]] = None,
            generator: Optional[Callable[
                [FeatureGenerator[T]], Union[T, Tuple[T, ...]]]] = None,
            randomized: Optional[bool] = None, allow_missing: bool = False):
        return cls(
            name=name, is_numerical=False,
            categories=categories, probabilities=probabilities,
            generator=generator, randomized=randomized,
            allow_missing=allow_missing)

    def __post_init__(self):
        if self.is_numerical:
            if self.categories is not None:
                raise ValueError('Numerical type cannot have categories.')
            if self.allow_missing:
                raise TypeError('Only categorical type can accept '
                                'missing values.')

            self._process_numerical()
        else:
            if self.lower is not None or self.upper is not None:
                raise ValueError(
                    'Categorical type cannot have lower and upper.')

            probabilities = self.probabilities
            categories = self.categories
            if probabilities is not None:
                if abs(sum(probabilities) - 1.0) > 1e-6:
                    raise ValueError('probabilities should add up to 1.0.'
                                     f'Got {sum(probabilities)}')
                if categories is None:
                    raise ValueError(
                        'probabilities cannot be set for None categories.')
                if len(probabilities) != len(categories):
                    raise ValueError('Size mismatch. '
                                     f'{len(categories)} categories vs. '
                                     f'{len(probabilities)} probabilities')

            self._process_categorical()

    def __call__(
        self, value: Optional[Union[T, Tuple[T, ...], MissingType]] = None
    ) -> Union[Feature[T], Feature[Tuple[T, ...]]]:
        if value is None:
            if self.generator is None:
                raise RuntimeError('generator not found.')
            _value = self.generator(self)
        else:
            _value = value

        if self.is_numerical:
            if _value == MISSING:
                raise ValueError('Numerical feature cannot accept MISSING.')

            return self._call_numerical(_value)
        else:
            return self._call_categorical(_value)

    def _process_categorical(self):
        if self.categories is None:
            return

        allow_missing_offset = int(self.allow_missing)
        cat_count = len(self.categories) - 1 + allow_missing_offset
        normalizer = {}
        for i, c in enumerate(self.categories[:-1]):
            temp = [0] * cat_count
            temp[i] = 1
            normalizer[c] = tuple(temp)

        temp = [0] * cat_count
        temp[-1] = allow_missing_offset
        normalizer[self.categories[-1]] = tuple(temp)

        if self.allow_missing:
            normalizer[MISSING] = tuple([0] * cat_count)

        self.__dict__['normalizer'] = normalizer

    def _process_numerical(self):
        lower: Optional[float] = self.lower  # type: ignore
        upper: Optional[float] = self.upper  # type: ignore

        if lower is None or upper is None or upper == lower:
            self.__dict__['normalizer'] = lambda _: None  # type: ignore
        else:
            if lower > upper:
                raise ValueError(f'lower ({lower}) cannot be '
                                 f'greater than upper ({upper}).')

            denominator = upper - lower

            def normalizer(x: float) -> float:
                return (x - lower)/denominator  # type: ignore

            self.__dict__['normalizer'] = normalizer

    def _call_categorical(
            self, value: Union[T, Tuple[T, ...], MissingType]
    ) -> Union[Feature[T], Feature[Tuple[T, ...]]]:
        normalizer = self.normalizer
        categories = self.categories or ()

        if normalizer is None:
            normalized = None
        elif value in categories:
            normalized = normalizer[value]
        elif self.allow_missing and value == MISSING:
            normalized = normalizer[value]
        else:
            try:
                normalized = tuple(
                    itertools.chain(
                        *(normalizer[d]
                          for d in value)))  # type: ignore
            except KeyError:
                raise ValueError(
                    f'{value} is not '
                    f'in the categories={categories}.')

        instance = Feature.categorical(
            name=self.name, value=value, categories=self.categories,
            normalized=normalized)

        return instance

    def _call_numerical(
            self, value: Union[T, Tuple[T, ...]]
    ) -> Union[Feature[T], Feature[Tuple[T, ...]]]:
        normalizer = self.normalizer
        lower = self.lower
        upper = self.upper

        if value is None:
            normalized = None

        elif isinstance(value, tuple):
            if (lower is not None
                    and min(value) < lower):  # type: ignore
                raise ValueError(f'Lower bound ({lower}) violated:\n {value}')

            if (upper is not None
                    and max(value) > upper):  # type: ignore
                raise ValueError(f'Upper bound ({upper}) violated:\n {value}')

            normalized = tuple(normalizer(d) for d in value)  # type: ignore

        else:
            if (lower is not None
                    and value < lower):  # type: ignore
                raise ValueError(f'Lower bound ({lower}) violated:\n {value}')

            if (upper is not None
                    and value > upper):  # type: ignore
                raise ValueError(f'Upper bound ({upper}) violated:\n {value}')

            normalized = normalizer(value)  # type: ignore

        instance: Union[Feature[T], Feature[Tuple[T, ...]]] = \
            Feature.numerical(name=self.name, value=value,   # type: ignore
                              lower=lower, upper=upper,
                              normalized=normalized)

        return instance


class FeatureArray:
    '''
    The main datatype used to communicate `state`s, `action`s, and `reward`s,
    between objects in `reil`.
    '''

    def __init__(self, data: Union[Feature[Any], Iterable[Feature[Any]]]):
        '''
        Arguments
        ---------
        data:
            One or a sequence of `Feature`s.
        '''
        temp: Dict[str, Feature[Any]] = {}
        _data: Iterable[Any] = (  # type: ignore
            data if hasattr(data, '__iter__') else [data])

        for d in _data:
            if isinstance(d, Feature):
                name = d.name
                if name in temp:
                    raise KeyError(f'Duplicate name ({name}).')

                temp[name] = d
            else:
                raise TypeError(f'Unknown input type {type(d)} for item: {d}')

        self._data = temp
        self._clear_temps()

    def _clear_temps(self):
        self._value: Union[Dict[str, Any], None] = None
        self._lower: Union[Dict[str, Any], None] = None
        self._upper: Union[Dict[str, Any], None] = None
        self._categories: Union[Dict[str, Any], None] = None
        self._is_numerical: Union[Dict[str, Any], None] = None

    @property
    def value(self):
        '''
        Return a dictionary with elements' names as keys and
        their respective values as values.

        Returns
        -------
        :
            Names of the elements and their values.
        '''
        if self._value is None:
            self._value = {name: v.value
                           for name, v in self._data.items()}

        return self._value

    @property
    def lower(self):
        '''
        Return all `lower` attributes.

        Returns
        -------
        :
            `lower` attribute of all `NumericalData` variables with their names
            as keys.
        '''
        if self._lower is None:
            self._lower = {name: v.lower
                           for name, v in self._data.items()}

        return self._lower

    @property
    def upper(self):
        '''
        Return all `upper` attributes.

        Returns
        -------
        :
            `upper` attribute of all `NumericalData` variables with their names
            as keys.
        '''
        if self._upper is None:
            self._upper = {name: v.upper
                           for name, v in self._data.items()}

        return self._upper

    @property
    def categories(self):
        '''
        Return all `categories` attributes.

        Returns
        -------
        :
            `categories` attribute of all `CategoricalData` variables with
            their names as keys.
        '''
        if self._categories is None:
            self._categories = {name:
                                v.categories
                                for name, v in self._data.items()}

        return self._categories

    @property
    def normalized(self):
        '''
        Normalize all items in the instance.

        Returns
        -------
        :
            A `FeatureArray` of the normalized values of all the items in the
            instance, in the form of numerical `Feature`s.
        '''
        return FeatureArray(
            FeatureGenerator.numerical(name=name, lower=0, upper=1)
            (v.normalized)  # type: ignore
            for name, v in self._data.items())

    def flatten(self):
        """Combine values of all items in the instance.

        Returns
        -------
        :
            A list that contains all the values of all the items.
        """
        def make_iterable(x: Any) -> Iterable[Any]:
            return x if hasattr(x, '__iter__') else [x]

        return list(itertools.chain(*[make_iterable(sublist)
                                      for sublist in self.value.values()]))

    def split(self):
        """Split the `FeatureArray` into a list of `FeatureArray`s.

        Returns
        -------
        :
            All items in the instance as separate `FeatureArray` instances.
        """
        if len(self) == 1:
            d = next(iter(self._data.values()))
            if not isinstance(d.value, (list, tuple)):
                splitted_list = FeatureArray(d)
            else:
                temp = d.as_dict()
                cls = type(d)
                value = temp['value']
                del temp['value']
                if 'is_numerical' in temp:
                    del temp['is_numerical']

                splitted_list = [
                    FeatureArray(cls(
                        value=v,
                        **temp))
                    for v in value]

        else:
            splitted_list = list(FeatureArray(v) for v in self._data.values())

        return splitted_list

    def __iter__(self):
        return iter(self._data.values())

    def __getitem__(self, k: str):
        return self._data.__getitem__(k)

    def __len__(self):
        return self._data.__len__()

    def __hash__(self):
        return hash(tuple(self._data.items()))

    def __eq__(self, other: Any):
        return isinstance(other, type(self)) and (self._data == other._data)

    def __add__(self, other: Any):
        if not isinstance(other, FeatureArray):
            new_data = FeatureArray(other)
        else:
            new_data = other

        # if not isinstance(new_data, FeatureArray):
        #     raise TypeError(
        #         'Concatenation of type FeatureArray'
        #         f' and {type(other)} not implemented!')

        overlaps = set(new_data._data).intersection(self._data)
        if overlaps:
            raise ValueError(f'Objects already exist: {overlaps}.')

        return FeatureArray(itertools.chain(self._data.values(),
                                            new_data._data.values()))

    def __neg__(self):
        temp = [v
                for v in self._data.values()]
        for item in temp:
            if hasattr(item.value, '__neg__'):
                neg_value = -item.value  # type: ignore
                lower: Any = item.__dict__.get('lower') or neg_value
                upper: Any = item.__dict__.get('upper') or neg_value
                if lower <= neg_value <= upper:
                    object.__setattr__(item, 'value', neg_value)
                else:
                    raise ValueError(f'Bounds violated: lower: {lower}, '
                                     f'upper: {upper}, '
                                     f'negative value: {neg_value}')

        return FeatureArray(temp)  # type: ignore

    def __repr__(self):
        return f'[{super().__repr__()} -> {self._data}]'

    def __str__(self):
        return f"[{', '.join((d.__str__() for d in self._data.items()))}]"


def change_to_missing(feature: Feature[T]) -> Feature[T]:
    if feature.is_numerical:
        raise TypeError('Only categorical features can have missing.')
    categories = feature.categories
    normalized = feature.normalized
    if categories is None:
        raise ValueError('No categories defined!')
    if normalized is None:
        raise ValueError('Cannot generate normal form for a feature '
                         'without the normal form.')

    if len(categories) != len(normalized):
        raise TypeError('Feature is not allowed to have MISSING')

    return FeatureGenerator.categorical(  # type: ignore
        name=feature.name, categories=categories, allow_missing=True)(MISSING)


def change_array_to_missing(
        features: FeatureArray, suppress_error: bool = True) -> FeatureArray:
    def try_to_change(feature: Feature[T]) -> Feature[T]:
        try:
            return change_to_missing(feature)
        except (TypeError, ValueError):
            if suppress_error:
                return feature
            raise

    return FeatureArray(try_to_change(f) for f in features)



# -*- coding: utf-8 -*-
'''
PrimaryComponent and SecondayComponent classes
==============================================

A datatype used to specify entity components, such as `state`, `reward`,
and `statistic`.
'''
from __future__ import annotations

import dataclasses
from collections import defaultdict
from typing import (Any, Callable, DefaultDict, Dict, Generic, List, Optional,
                    Tuple, TypeVar, Union)

import pandas as pd
from reil.datatypes.feature import FeatureArray

SubComponentInfo = Tuple[Callable[..., Dict[str, Any]], Tuple[str, ...]]

ArgsType = TypeVar('ArgsType', str, Tuple[str, ...], Dict[str, Any])


@dataclasses.dataclass
class SubComponentInstance(Generic[ArgsType]):
    '''
    A `dataclass` to store an instance of a sub component.

    :meta private:
    '''
    name: str
    args: ArgsType
    fn: Callable[..., Any]


class PrimaryComponent:
    '''
    The datatype to specify primary component, e.g., `state`.
    '''

    def __init__(
        self,
        object_ref: object,
        available_sub_components: Optional[Dict[str, SubComponentInfo]] = None,
        default_definition: Optional[Callable[[
            Optional[int]], FeatureArray]] = None
    ) -> None:
        '''
        Parameters
        ----------
        available_sub_components:
            A dictionary with sub component names as keys and a tuple of
            function and its argument list as values.
        '''
        self._available_sub_components: Dict[str, SubComponentInfo] = {}
        self._definitions: Dict[str, List[
            SubComponentInstance[Dict[str, Any]]]] = defaultdict(list)

        if available_sub_components is not None:
            self.sub_components = available_sub_components

        self.object_ref = object_ref
        self._default = default_definition

    @property
    def sub_components(self) -> Dict[str, SubComponentInfo]:
        '''Get and set the dictionary of sub components.

        Returns
        -------
        :
            Sub components

        Notes
        -----
        Sub components info can only be set once.
        '''
        return self._available_sub_components

    @sub_components.setter
    def sub_components(self, sub_components: Dict[str, SubComponentInfo]):
        if self._available_sub_components:
            raise ValueError('Available sub components list is already set. '
                             'Cannot modify it.')
        self._available_sub_components = sub_components

    def set_default_definition(
            self, default_definition: Callable[[Optional[int]], FeatureArray]
    ) -> None:
        '''Add a new component definition.

        Parameters
        ----------
        default_definition:
            A function that can optionally accept `_id`, and returns a
            `FeatureArray`.
        '''
        self._default = default_definition

    def add_definition(
            self,
            name: str,
            *sub_components: Tuple[str, Dict[str, Any]]) -> None:
        '''Add a new component definition.

        Parameters
        ----------
        name:
            The name of the new component.

        sub_components:
            Sub components that form this new component. Each sub component
            should be specified as a tuple. The first item is the name of the
            sub component, and the second item is a dictionary of kwargs and
            values for that sub component.

        Raises
        ------
        ValueError
            Definition already exists for this name.

        ValueError
            Unknown sub component.

        ValueError
            Unknown keyword argument.
        '''
        _name = name.lower()

        if _name == 'default':
            raise ValueError('Use `set_default_definition` for the default '
                             'definition')

        if _name in self._definitions:
            raise ValueError(f'Definition {name} already exists.')

        unknown_sub_components = set(
            sc for sc, _ in sub_components).difference(
            self._available_sub_components)

        if unknown_sub_components:
            raise ValueError('Unknown sub components: '
                             f'{unknown_sub_components}')

        for sub_comp_name, kwargs in sub_components:
            fn, arg_list = self._available_sub_components[sub_comp_name]

            unknown_keywords = set(kwargs).difference(arg_list)
            if unknown_keywords:
                raise ValueError(
                    f'Unknown keyword argument(s): {unknown_keywords}.')

            self._definitions[_name].append(SubComponentInstance(
                name=sub_comp_name, fn=fn, args=kwargs))

    @property
    def definitions(self):
        '''Return the dictionary of component definitions.

        Returns
        -------
        :
            The dictionary of component definitions.
        '''
        return self._definitions

    def default(self, _id: Optional[int] = None) -> FeatureArray:
        '''
        Generate the default component definition.

        Parameters
        ----------
        _id:
            ID of the caller object

        Returns
        -------
        :
            The component with the default definition.
        '''
        if self._default is None:
            raise AttributeError('Default definition not found.')

        return self._default(_id)

    def __call__(self, name: str, _id: Optional[int] = None) -> FeatureArray:
        '''
        Generate the component based on the specified `name` for the
        specified caller.

        Parameters
        ----------
        name:
            The name of the component definition.

        _id:
            ID of the caller.

        Returns
        -------
        :
            The component with the specified definition `name`.

        Raises
        ------
        ValueError
            Definition not found.
        '''
        if name == 'default':
            try:
                return self.default(_id)
            except AttributeError:
                pass

        if name not in self._definitions:
            raise ValueError(f'Definition {name} not found.')

        return FeatureArray(d.fn(
            self.object_ref, _id=_id, **d.args)  # type: ignore
            for d in self._definitions[name.lower()])


class SecondayComponent:
    '''
    The datatype to specify secondary components, e.g. `statistic` and
    `reward`.
    '''

    def __init__(
            self,
            name: str,
            primary_component: Optional[PrimaryComponent] = None,
            default_definition: Optional[Callable[[
                Optional[int]], Any]] = None,
            enabled: bool = True):
        '''

        Parameters
        ----------
        name:
            The name of the secondary component.

        primary_component:
            An instance of a `PrimaryComponent` from which component
            definitions are used.

        default_definition:
            The `default` definition.

        enabled:
            Whether to return the computed value or `None`.
        '''
        self._name = name
        self._primary_component = primary_component
        self._default = default_definition
        self._enabled = enabled

        self._definitions: Dict[
            str, SubComponentInstance[str]] = defaultdict(None)

    def enable(self) -> None:
        self._enabled = True

    def disable(self) -> None:
        self._enabled = False

    def set_primary_component(
            self,
            primary_component: PrimaryComponent) -> None:
        '''Set the primary component.

        Parameters
        ----------
        primary_component:
            An instance of a `PrimaryComponent` from which component
            definitions are used.

        Raises
        ------
        ValueError
            Primary component is already set.
        '''
        if self._primary_component is not None:
            raise ValueError('Primary component is already set. '
                             'Cannot modify it.')

        self._primary_component = primary_component

    def set_default_definition(
            self, default_definition: Callable[[Optional[int]], Any]
    ) -> None:
        '''Add a new component definition.

        Parameters
        ----------
        default_definition:
            A function that can optionally accept `_id`, and returns a value.
        '''
        self._default = default_definition

    def add_definition(
            self, name: str, fn: Callable[..., Any],
            primary_component_name: str = 'default') -> None:
        '''
        Add a new component definition.

        Parameters
        ----------
        name:
            The name of the new component.

        fn:
            The function that will receive the primary component instance and
            computes the value of the secondary component.

        primary_component_name:
            The component name that will be used by `fn`.

        Raises
        ------
        ValueError
            Definition already exists for this name.

        ValueError
            Undefined primary component name.
        '''
        _name = name.lower()
        _primary_component_name = primary_component_name.lower()

        if _name == 'default':
            raise ValueError('Use `set_default_definition` for the default '
                             'definition')

        if _name in self._definitions:
            raise ValueError(f'Definition {name} already exists.')

        if self._primary_component is None:
            raise ValueError(
                'Primary component is not defined. '
                'Use `set_primary_component` to specify it.')

        if _primary_component_name not in self._primary_component.definitions:
            raise ValueError(f'Undefined {_primary_component_name}.')

        self._definitions[_name] = SubComponentInstance(
            name=_name,
            fn=fn,
            args=_primary_component_name)

    def default(self, _id: Optional[int] = None) -> Any:
        '''
        Generate the default component definition.

        Parameters
        ----------
        _id:
            ID of the caller object

        Returns
        -------
        :
            The component with the default definition.
        '''
        if self._default is not None:
            return self._default(_id)

        raise AttributeError('Default definition not found.')

    def __call__(self,
                 name: str,
                 _id: Optional[int] = None) -> Any:
        '''
        Generate the component based on the specified `name` for the
        specified caller.

        Parameters
        ----------
        name:
            The name of the component definition.

        _id:
            ID of the caller.

        Returns
        -------
        :
            The component with the specified definition `name`.

        Raises
        ------
        ValueError
            Definition not found.
        '''
        if not self._enabled:
            return None

        _name = name.lower()

        if _name == 'default':
            try:
                return self.default(_id)
            except AttributeError:
                pass

        if self._primary_component is None:
            raise ValueError(
                'Primary component is not defined. '
                'Use `set_primary_component` to specify it.')

        try:
            d = self._definitions[_name]
        except KeyError:
            raise ValueError(f'Definition {name} not found.')

        p = self._primary_component(name=d.args, _id=_id)

        return d.fn(p)


class Statistic:
    '''
    A component similar to `SecondaryComponent`, but with history and
    aggregator.
    '''

    def __init__(
            self,
            name: str,
            primary_component: Optional[PrimaryComponent] = None,
            default_definition: Optional[Callable[[
                Optional[int]], Tuple[FeatureArray, float]]] = None,
            enabled: bool = True):
        '''

        Parameters
        ----------
        name:
            The name of the secondary component.

        primary_component:
            An instance of a `PrimaryComponent` from which component
            definitions are used.

        default_definition:
            The `default` definition.

        enabled:
            Whether to return the computed value or `None`.
        '''
        self._name = name
        self._primary_component = primary_component
        self._default = default_definition
        self._enabled = enabled

        self._definitions: Dict[
            str, SubComponentInstance[Tuple[str, str]]] = defaultdict(None)

        self._history: Dict[
            int,
            List[Tuple[FeatureArray, float]]] = DefaultDict(list)
        self._history_none: List[Tuple[FeatureArray, float]] = []

    def enable(self) -> None:
        self._enabled = True

    def disable(self) -> None:
        self._enabled = False

    def set_primary_component(
            self,
            primary_component: PrimaryComponent) -> None:
        '''Set the primary component.

        Parameters
        ----------
        primary_component:
            An instance of a `PrimaryComponent` from which component
            definitions are used.

        Raises
        ------
        ValueError
            Primary component is already set.
        '''
        if self._primary_component is not None:
            raise ValueError('Primary component is already set. '
                             'Cannot modify it.')

        self._primary_component = primary_component

    def set_default_definition(
            self,
            default_definition: Callable[
                [Optional[int]], Tuple[FeatureArray, float]]) -> None:
        '''Add a new component definition.

        Parameters
        ----------
        default_definition:
            A function that can optionally accept `_id`, and returns a
            `FeatureArray`.
        '''
        self._default = default_definition

    def add_definition(
            self, name: str, fn: Callable[..., Any],
            stat_component: str, aggregation_component: str) -> None:
        '''
        Add a new component definition.

        Parameters
        ----------
        name:
            The name of the new component.

        fn:
            The function that will receive the primary component instance and
            computes the value of the secondary component.

        stat_component:
            The component name that will be used by `fn`.

        aggregation_component:
            The component name that will be used to do aggregation.

        Raises
        ------
        ValueError
            Definition already exists for this name.

        ValueError
            Undefined primary component name.
        '''
        _name = name.lower()
        _stat_component = stat_component.lower()
        _aggregation_component = aggregation_component.lower()
        if _name == 'default':
            raise ValueError('Use `set_default_definition` for the default '
                             'definition')

        if _name in self._definitions:
            raise ValueError(f'Definition {name} already exists.')

        if self._primary_component is None:
            raise ValueError(
                'Primary component is not defined. '
                'Use `set_primary_component` to specify it.')

        if _stat_component not in self._primary_component.definitions:
            raise ValueError(f'Undefined {_stat_component}.')

        if _aggregation_component not in self._primary_component.definitions:
            raise ValueError(f'Undefined {_aggregation_component}.')

        self._definitions[_name] = SubComponentInstance[Tuple[str, str]](
            name=_name,
            fn=fn,
            args=(_aggregation_component, _stat_component))

    def default(self, _id: Optional[int] = None) -> Tuple[FeatureArray, float]:
        '''
        Generate the default component definition.

        Parameters
        ----------
        _id:
            ID of the caller object

        Returns
        -------
        :
            The component with the default definition.
        '''
        if self._default is not None:
            return self._default(_id)

        raise AttributeError('Default definition not found.')

    def __call__(
            self,
            name: str,
            _id: Optional[int] = None
    ) -> Union[Tuple[FeatureArray, float], None]:
        '''
        Generate the component based on the specified `name` for the
        specified caller.

        Parameters
        ----------
        name:
            The name of the component definition.

        _id:
            ID of the caller.

        Returns
        -------
        :
            The component with the specified definition `name`.

        Raises
        ------
        ValueError
            Definition not found.
        '''
        if not self._enabled:
            return None

        _name = name.lower()

        if _name == 'default':
            try:
                return self.default(_id)
            except AttributeError:
                pass

        if self._primary_component is None:
            raise ValueError(
                'Primary component is not defined. '
                'Use `set_primary_component` to specify it.')

        try:
            d = self._definitions[_name]
        except KeyError:
            raise ValueError(f'Definition {name} not found.')

        agg, comp_name = d.args

        return (self._primary_component(name=agg, _id=_id),
                d.fn(self._primary_component(name=comp_name, _id=_id)))

    def append(self,
               name: str,
               _id: Optional[int] = None) -> None:
        '''
        Generate the stat and append it to the history.

        Arguments
        ---------
        name:
            The name of the component definition.

        _id:
            ID of the caller.

        Raises
        ------
        ValueError
            Definition not found.
        '''
        s = self.__call__(name, _id)
        if s is not None:
            if _id is None:
                self._history_none.append(s)
            else:
                self._history[_id].append(s)

    def aggregate(self,
                  aggregators: Optional[Tuple[str, ...]] = None,
                  groupby: Optional[Tuple[str, ...]] = None,
                  _id: Optional[int] = None,
                  reset_history: bool = False):
        temp = self._history_none if _id is None else self._history[_id]
        if not temp:
            return None

        df = pd.DataFrame({'instance_id': i,  # type: ignore
                           **x[0].value,
                           'value': x[1]}
                          for i, x in enumerate(temp))
        temp_group_by = ['instance_id'] if groupby is None else list(groupby)
        grouped_df = df.groupby(temp_group_by)

        def no_change(x: Any) -> Any:
            return x

        result: pd.DataFrame = grouped_df['value'].agg(  # type: ignore
            aggregators or no_change)

        if reset_history:
            self._history: Dict[
                int, List[Tuple[FeatureArray, float]]] = DefaultDict(list)
            self._history_none: List[Tuple[FeatureArray, float]] = []

        return result



# -*- coding: utf-8 -*-
'''
AAA, CAA, PGAA, PGPGA, PGPGI classes
====================================

Study arms in `Ravvaz et al. (2017)
<https://doi.org/10.1161/circgenetics.117.001804>`_
'''

from typing import Any, Dict

from reil.healthcare.dosing_protocols.dosing_protocol import DosingDecision
from reil.healthcare.dosing_protocols.three_phase_dosing_protocol import \
    ThreePhaseDosingProtocol
from reil.healthcare.dosing_protocols.warfarin.aurora import Aurora
from reil.healthcare.dosing_protocols.warfarin.intermountain import \
    Intermountain
from reil.healthcare.dosing_protocols.warfarin.iwpc import IWPC
from reil.healthcare.dosing_protocols.warfarin.lenzini import Lenzini


class AAA(ThreePhaseDosingProtocol):
    '''
    A composite dosing protocol with `Aurora` in all phases.
    '''

    def __init__(self) -> None:
        aurora_instance = Aurora()
        super().__init__(aurora_instance, aurora_instance, aurora_instance)

    def prescribe(
            self, patient: Dict[str, Any]) -> DosingDecision:
        dosing_decision, self._additional_info = \
            self._initial_protocol.prescribe(patient, self._additional_info)

        return dosing_decision

    def __repr__(self) -> str:
        return super().__repr__() + '[AAA]'


class CAA(ThreePhaseDosingProtocol):
    '''
    A composite dosing protocol with clinical `IWPC` in the initial phase
    (days 1 and 2), and `Aurora` in the adjustment and maintenance phases.
    '''

    def __init__(self) -> None:
        iwpc_instance = IWPC('clinical')
        aurora_instance = Aurora()
        super().__init__(iwpc_instance, aurora_instance, aurora_instance)

    def prescribe(
            self, patient: Dict[str, Any]) -> DosingDecision:
        day: int = patient['day']
        if day <= 2:
            temp, self._additional_info = self._initial_protocol.prescribe(
                patient, self._additional_info)
            dosing_decision = DosingDecision(temp.dose, 3 - day)
        else:
            dosing_decision, self._additional_info = \
                self._adjustment_protocol.prescribe(
                    patient, self._additional_info)

        return dosing_decision

    def __repr__(self) -> str:
        return super().__repr__() + '[CAA]'


class PGAA(ThreePhaseDosingProtocol):
    '''
    A composite dosing protocol with pharmacogenetic `IWPC` in the initial
    phase (days 1 and 2), and `Aurora` in the adjustment and maintenance
    phases.
    '''

    def __init__(self) -> None:
        iwpc_instance = IWPC('pharmacogenetic')
        aurora_instance = Aurora()
        super().__init__(iwpc_instance, aurora_instance, aurora_instance)

    def prescribe(
            self, patient: Dict[str, Any]) -> DosingDecision:
        day: int = patient['day']
        if day <= 2:
            temp, self._additional_info = self._initial_protocol.prescribe(
                patient, self._additional_info)
            dosing_decision = DosingDecision(temp.dose, 3 - day)
        else:
            dosing_decision, self._additional_info = \
                self._adjustment_protocol.prescribe(
                    patient, self._additional_info)

        return dosing_decision

    def __repr__(self) -> str:
        return super().__repr__() + '[PGAA]'


class PGPGA(ThreePhaseDosingProtocol):
    '''
    A composite dosing protocol with modified `IWPC` in the initial phase
    (days 1, 2, and 3), `Lenzini` in the adjustment phase (days 4 and 5),
    and `Aurora` in the maintenance phase.
    '''

    def __init__(self) -> None:
        iwpc_instance = IWPC('modified')
        lenzini_instance = Lenzini()
        aurora_instance = Aurora()
        super().__init__(iwpc_instance, lenzini_instance, aurora_instance)

    def prescribe(
            self, patient: Dict[str, Any]) -> DosingDecision:
        if patient['day'] <= 3:
            temp, self._additional_info = \
                self._initial_protocol.prescribe(
                    patient, self._additional_info)
            dosing_decision = DosingDecision(temp.dose, 4 - patient['day'])
        elif patient['day'] <= 5:
            temp, self._additional_info = \
                self._adjustment_protocol.prescribe(
                    patient, self._additional_info)
            dosing_decision = DosingDecision(temp.dose, 6 - patient['day'])
        else:
            dosing_decision, self._additional_info = \
                self._maintenance_protocol.prescribe(
                    patient, self._additional_info)

        return dosing_decision

    def __repr__(self) -> str:
        return super().__repr__() + '[PGPGA]'


class PGPGI(ThreePhaseDosingProtocol):
    '''
    A composite dosing protocol with modified `IWPC` in the initial phase
    (days 1, 2, and 3), `Lenzini` in the adjustment phase (days 4 and 5),
    and `Intermountain` in the maintenance phase.
    '''

    def __init__(self) -> None:
        iwpc_instance = IWPC('modified')
        lenzini_instance = Lenzini()
        intermountain_instance = Intermountain(enforce_day_ge_8=False)
        super().__init__(
            iwpc_instance, lenzini_instance, intermountain_instance)

    def prescribe(
            self, patient: Dict[str, Any]) -> DosingDecision:
        if patient['day'] <= 3:
            temp, self._additional_info = \
                self._initial_protocol.prescribe(
                    patient, self._additional_info)
            dosing_decision = DosingDecision(temp.dose, 4 - patient['day'])
        elif patient['day'] <= 5:
            temp, self._additional_info = \
                self._adjustment_protocol.prescribe(
                    patient, self._additional_info)
            dosing_decision = DosingDecision(temp.dose, 6 - patient['day'])
        else:
            dosing_decision, self._additional_info = \
                self._maintenance_protocol.prescribe(
                    patient, self._additional_info)

        return dosing_decision

    def __repr__(self) -> str:
        return super().__repr__() + '[PGPGI]'



# -*- coding: utf-8 -*-
'''
HealthMathModel class
=====================

The base class of all mathematical models, e.g. ODE, PDE, PK/PD, etc. in
healthcare.
'''
from typing import Any, Dict

from reil.datatypes.feature import Feature


class HealthMathModel:
    '''
    The base class of all mathematical models in healthcare.
    '''

    def setup(self, **arguments: Feature[Any]) -> None:
        '''
        Set up the model.

        Arguments
        ---------
        arguments:
            Any parameter that the model needs to setup initially.
        '''
        raise NotImplementedError

    def run(self, **inputs: Any) -> Dict[str, Any]:
        '''
        Run the model.

        Arguments
        ---------
        inputs:
            Any input arguments that the model needs for the run.

        Returns
        -------
        :
            A dictionary of model's return values.
        '''
        raise NotImplementedError



# -*- coding: utf-8 -*-
'''
InstanceGenerator class
=======================

`InstanceGenerator` takes any object derived form `ReilBase` and returns
an iterator.
'''

from __future__ import annotations

import os
import pathlib
from typing import Any, Generic, Tuple, TypeVar, Union

from reil import reilbase, stateful
from reil.datatypes.mock_statistic import MockStatistic

T = TypeVar('T', bound=stateful.Stateful)


class InstanceGenerator(Generic[T], reilbase.ReilBase):
    '''
    Make any ReilBase object an iterable.

    The initializer accepts, among other arguments, an instance of an object to
    iterate, and `instance_counter_stops`, which is a tuple of the instance
    numbers where the instance generator should stop.
    '''

    def __init__(
            self,
            object: T,
            instance_counter_stops: Tuple[int] = (-1,),  # -1: infinite
            first_instance_number: int = 0,
            auto_rewind: bool = False,
            save_instances: bool = False,
            overwrite_instances: bool = False,
            use_existing_instances: bool = True,
            save_path: Union[pathlib.PurePath, str] = '',
            filename_pattern: str = '{n:04}',
            **kwargs: Any):
        '''
        Attributes
        ----------
        object:
            An instance of an object.

        instance_counter_stops:
            A tuple of the instance numbers where the instance
            generator should stop. A value of -1 means infinite.

        first_instance_number:
            The number of the first instance to be generated.

        auto_rewind:
            Whether to rewind after the generator hits the last stop.

        save_instances:
            Whether to save instances of the `object` or not.

        overwrite_instances:
            Whether to overwrite instances of the `object` or not.
            This flag is useful only if `save_instances` is set to `True`.

        use_existing_instances:
            Whether try to load instances before attempting to create them.

        save_path:
            The path where instances should be saved to/ loaded from.

        filename_pattern:
            A string that uses "n" as the instance number, and is
            used for saving and loading instances.
        '''
        super().__init__(**kwargs)

        self._object = object
        self._save_instances = save_instances
        self._use_existing_instances = use_existing_instances
        self._overwrite_instances = overwrite_instances
        self._save_path = save_path
        self._filename_pattern = filename_pattern

        self._auto_rewind = auto_rewind
        self._first_instance_number = first_instance_number
        self._instance_counter_stops = instance_counter_stops
        self._last_stop_index = len(self._instance_counter_stops) - 1

        self.is_finite = -1 not in instance_counter_stops and not auto_rewind
        self.statistic = MockStatistic(self._object)

        self.rewind()

        if self._object is None:
            self._logger.info('Instance Generator initialized with None.')
        else:
            self._object._name = self._filename_pattern.format(
                n=self._instance_counter)

    @classmethod
    def _empty_instance(cls):
        return cls(None)

    def __iter__(self):
        return self

    def __next__(self) -> Tuple[int, T]:
        end = self._get_end()

        self._instance_counter += 1

        if self._stop_check(self._instance_counter, end):
            self._instance_counter -= 1
            self._stops_index += 1

            if self._stops_index <= self._last_stop_index:
                self._determine_stop_check()

            self._partially_terminated = True
            raise StopIteration

        else:
            self._partially_terminated = False
            self._generate_new_instance()

        return (self._instance_counter, self._object)

    def _get_end(self):
        try:
            end = self._instance_counter_stops[self._stops_index]
        except IndexError:
            if self._auto_rewind:
                self.rewind()
                end = self._instance_counter_stops[self._stops_index]
            else:
                raise StopIteration
        return end

    @staticmethod
    def _do_not_stop(current: int, end: int) -> bool:
        return False

    @staticmethod
    def _do_stop(current: int, end: int) -> bool:
        return current >= end

    def _determine_stop_check(self):
        if self._instance_counter_stops[self._stops_index] == -1:
            self._stop_check = self._do_not_stop
        else:
            self._stop_check = self._do_stop

    def _generate_new_instance(self):
        current_instance = self._filename_pattern.format(
                n=self._instance_counter)
        new_instance = True
        if self._use_existing_instances:
            try:
                self._object.load(path=self._save_path,
                                  filename=current_instance)
                new_instance = False
            except FileNotFoundError:
                self._object.reset()
        else:
            self._object.reset()

        self._object._name = current_instance
        self.statistic.set_object(self._object)

        if self._save_instances and new_instance:
            if (not self._overwrite_instances and
                    os.path.isfile(pathlib.Path(
                        self._save_path, f'{current_instance}.pkl'))):
                if self._use_existing_instances:
                    # This extra test is useful when multiple instances
                    # generators are active. One might create the instance and
                    # save it, while others are trying to create it. In such
                    # case, other generators raise exception, but should not!
                    self._object.load(path=self._save_path,
                                      filename=current_instance)
                else:
                    raise FileExistsError(
                            f'File {current_instance} already exists.')
            else:
                self._object.save(path=self._save_path,
                                  filename=current_instance)

    def rewind(self) -> None:
        '''
        Rewind the iterator object.
        '''
        self._instance_counter = self._first_instance_number - 1
        self._stops_index = 0
        self._partially_terminated = False
        self._determine_stop_check()

    def is_terminated(self, fully: bool = True) -> bool:
        if fully:
            return not self._auto_rewind and (self._stops_index >
                                              self._last_stop_index)
        else:
            return self._partially_terminated

    def __repr__(self) -> str:
        try:
            return (f'{self.__class__.__qualname__} '
                    f'-- {self._instance_counter} --> '
                    f'{self._object.__repr__()}')
        except AttributeError:
            return self.__class__.__qualname__



# from __future__ import annotations

import multiprocessing
import pathlib
from multiprocessing.context import BaseContext
from typing import Any, Dict, List, Literal, Optional, Union

from reil.agents.agent import Agent
from reil.agents.agent_demon import AgentDemon
from reil.environments.task import Task
from reil.environments.environment_static_map import EnvironmentStaticMap
from reil.subjects.subject_demon import SubjectDemon
from reil.subjects.subject import Subject

# TODO: Documentation


class Session:
    def __init__(
            self, name: str, path: Union[pathlib.PurePath, str],
            main_task: Task,
            agents: Dict[str, Union[Agent[Any], str]],
            subjects: Dict[str, Union[Subject, str]],
            demons: Optional[Dict[
                str, Union[AgentDemon[Any], SubjectDemon, str]]] = None,
            tasks_before: Optional[List[Task]] = None,
            tasks_after: Optional[List[Task]] = None,
            tasks_before_iteration: Optional[List[Task]] = None,
            tasks_after_iteration: Optional[List[Task]] = None,
            separate_process: Optional[List[Literal[
                'tasks_before',
                'tasks_after',
                'tasks_before_iteration',
                'tasks_after_iteration']]] = None,
            process_type: Optional[Literal[
                'spawn', 'fork', 'forkserver']] = None):

        self._environment = EnvironmentStaticMap(
            name=name, path=pathlib.PurePath(path), demon_dict=demons or {})
        self._environment.add_entities(agents)  # type: ignore
        self._environment.add_entities(subjects)  # type: ignore

        self._main_task = main_task
        self._tasks_before = tasks_before
        self._tasks_after = tasks_after
        self._tasks_before_iteration = tasks_before_iteration
        self._tasks_after_iteration = tasks_after_iteration

        self._separate_process = separate_process or []
        if separate_process:
            self._process_type = process_type or 'spawn'
        else:
            self._process_type = None

    @staticmethod
    def _run_tasks(
            task_list: Optional[List[Task]],
            environment: EnvironmentStaticMap,
            iteration: int, separate_process: bool,
            context: Optional[BaseContext] = None):
        path = environment._path / environment._name
        p = None
        if task_list:
            for t in task_list:
                filename = f'{t._name}_{str(iteration)}'
                environment.interaction_sequence = ()
                if separate_process and context:
                    environment.save(
                        filename=filename, path=path)
                    p = context.Process(
                        target=t.run_file, args=(filename, path, iteration))
                    p.start()
                else:
                    t.run_env(environment, iteration)

    def run(self):
        '''Run the session'''
        context = (multiprocessing.get_context(self._process_type)
                   if self._separate_process else None)

        self._run_tasks(
            task_list=self._tasks_before,
            environment=self._environment,
            iteration=0,
            separate_process='tasks_before' in self._separate_process,
            context=context)

        itr = 0
        for itr in range(itr, self._main_task.max_iterations):
            self._run_tasks(
                task_list=self._tasks_before_iteration,
                environment=self._environment,
                iteration=itr,
                separate_process=(
                    'tasks_before_iteration' in self._separate_process),
                context=context)

            # path = self._environment._path / self._environment._name
            # filename = f'{self._environment._name}_{str(itr)}'
            # self._environment.save(
            #     filename=filename, path=path)
            self._environment = self._main_task.run_env(
                self._environment, itr)

            self._run_tasks(
                task_list=self._tasks_after_iteration,
                environment=self._environment,
                iteration=itr,
                separate_process=(
                    'tasks_after_iteration' in self._separate_process),
                context=context)

        self._run_tasks(
            task_list=self._tasks_after,
            environment=self._environment,
            iteration=itr + 1,
            separate_process='tasks_after' in self._separate_process,
            context=context)

        path = self._environment._path / self._environment._name
        filename = (f'{self._environment._name}_'
                    + str(self._main_task.max_iterations))
        self._environment.save(
            filename=filename, path=path)



# -*- coding: utf-8 -*-
'''
ActionGenerator class
=====================

Gets lists of categorical or numerical lists as components, and generates lists
of `FeatureArray` objects using the product of these components.
'''
import dataclasses
import itertools
from typing import (Any, Dict, Generic, Iterator, Optional, Tuple, TypeVar,
                    Union)

from reil.datatypes.feature import Feature, FeatureArray, FeatureGenerator
from reil.reilbase import ReilBase

Categorical = TypeVar('Categorical')
Numerical = TypeVar('Numerical', int, float)


@dataclasses.dataclass(frozen=True)
class CategoricalComponent(Generic[Categorical]):
    name: str
    possible_values: Tuple[Tuple[Categorical, ...], ...]
    categories: Tuple[Categorical, ...]
    length: int = dataclasses.field(init=False, hash=False, compare=False)
    feature_generator: FeatureGenerator[Categorical] = dataclasses.field(
        init=False, hash=False, compare=False)

    def __post_init__(self):
        object.__setattr__(self, 'length', len(self.possible_values))
        object.__setattr__(
            self, 'feature_generator',
            FeatureGenerator.categorical(
                name=self.name, categories=self.categories))

    def generate(self, index: int) -> Iterator[
            Union[Feature[Categorical], Feature[Tuple[Categorical, ...]]]]:
        _index = min(index, self.length - 1)
        return (self.feature_generator(vi)
                for vi in self.possible_values[_index])


@dataclasses.dataclass(frozen=True)
class NumericalComponent(Generic[Numerical]):
    name: str
    possible_values: Tuple[Tuple[Numerical, ...], ...]
    lower: Numerical
    upper: Numerical
    length: int = dataclasses.field(init=False, hash=False, compare=False)
    feature_generator: FeatureGenerator[Numerical] = dataclasses.field(
        init=False, hash=False, compare=False)

    def __post_init__(self):
        object.__setattr__(self, 'length', len(self.possible_values))
        object.__setattr__(
            self, 'feature_generator',
            FeatureGenerator.numerical(
                name=self.name, lower=self.lower, upper=self.upper))

    def generate(self, index: int) -> Iterator[
            Union[Feature[Numerical], Feature[Tuple[Numerical, ...]]]]:
        _index = min(index, self.length - 1)
        return (self.feature_generator(vi)
                for vi in self.possible_values[_index])


class ActionGenerator(ReilBase, Generic[Categorical, Numerical]):
    '''
    Gets lists of categorical or numerical lists as components, and generates
    lists of `FeatureArray` objects using the product of these components.
    '''

    def __init__(
            self,
            components: Optional[Dict[
                str, Union[
                    CategoricalComponent[Categorical],
                    NumericalComponent[Numerical]]]] = None,
            **kwargs: Any) -> None:
        '''
        Initializes the `ActionGenerator` instance.
        '''
        super().__init__(**kwargs)

        if components is not None:
            self._components = components
        else:
            self._components: Dict[
                str, Union[
                    CategoricalComponent[Categorical],
                    NumericalComponent[Numerical]]] = {}
        self._max_index: int = 0
        self.reset()

    def add_categorical(
            self,
            component_name: str,
            possible_values: Tuple[Tuple[Categorical, ...], ...],
            categories: Tuple[Categorical, ...]) -> None:
        '''
        Add a categorical component.

        Arguments
        ---------
        component_name:
            Name of the component.

        possible_values:
            A list of lists of categorical values.

        categories:
            A list of all possible categories.

        Raises
        ------
        KeyError
            `component_name` is duplicate.

        Example
        -------
        >>> AG = ActionGenerator()
        >>> AG.add_categorical(
        ...     component_name='compass_directions',
        ...     possible_values=(('N',), ('E', 'W'), ('N', 'S')),
        ...     categories=('N', 'S', 'E', 'W'))
        '''
        if component_name in self._components:
            raise KeyError(f'Key {component_name} already exists.')

        self._components[component_name] = CategoricalComponent(
            component_name, possible_values, categories)

        self._max_index = max(
            self._max_index,
            len(self._components[component_name].possible_values))

    def add_numerical(
            self,
            component_name: str,
            possible_values: Tuple[Tuple[Numerical, ...]],
            lower: Numerical,
            upper: Numerical) -> None:
        '''
        Add a numerical component.

        Arguments
        ---------
        component_name:
            Name of the component.

        possible_values:
            A list of lists of numerical values.

        lower:
            The minimum value possible.

        upper:
            The maximum value possible.

        Raises
        ------
        KeyError:
            `component_name` is duplicate.

        Example
        -------
        >>> AG = ActionGenerator()
        >>> AG.add_numerical(component_name='odds',
        ...     possible_values=((1,), (1, 3)),
        ...     lower=1, upper=9)
        '''
        if component_name in self._components:
            raise KeyError(f'Key {component_name} already exists.')

        self._components[component_name] = NumericalComponent(
            component_name, possible_values, lower, upper)

        self._max_index = max(
            self._max_index,
            len(self._components[component_name].possible_values))

    def possible_actions(
            self,
            state: Optional[FeatureArray] = None) -> Tuple[FeatureArray, ...]:
        '''
        Generate and return a list of possible actions.

        In this implementation, an `index` keeps track of where on the list it
        is on each component, and each call of this method generates the
        product of component values and returns the result as a list of
        `FeatureArray`. The `index` is incremented by 1 unit. The last list of
        a component is used if it is exhausted.

        Arguments
        ---------
        state:
            An optional argument that provides the generator with the current
            state of a subject. Subclasses of `ActionGenerator` can use this
            argument to generate tailored actions.

        Example
        -------
        >>> AG = ActionGenerator()
        >>> AG.add_categorical(
        ...     component_name='compass_directions',
        ...     possible_values=(('N',), ('E', 'W'), ('N', 'S')),
        ...     categories=('N', 'S', 'E', 'W'))
        >>> AG.add_numerical(component_name='odds',
        ...     possible_values=((1,), (1, 3)),
        ...     lower=1, upper=9)
        >>> for i in ('1st', '2nd', '3rd', '4th'):
        ...     print(f'calling possible_actions for the {i} time:')
        ...     for action in AG.possible_actions():
        ...         print(action.value)
        calling possible_actions for the 1st time:
        {'compass_directions': 'N', 'odds': 1}
        calling possible_actions for the 2nd time:
        {'compass_directions': 'E', 'odds': 1}
        {'compass_directions': 'E', 'odds': 3}
        {'compass_directions': 'W', 'odds': 1}
        {'compass_directions': 'W', 'odds': 3}
        calling possible_actions for the 3rd time:
        {'compass_directions': 'N', 'odds': 1}
        {'compass_directions': 'N', 'odds': 3}
        {'compass_directions': 'S', 'odds': 1}
        {'compass_directions': 'S', 'odds': 3}
        calling possible_actions for the 4th time:
        {'compass_directions': 'N', 'odds': 1}
        {'compass_directions': 'N', 'odds': 3}
        {'compass_directions': 'S', 'odds': 1}
        {'compass_directions': 'S', 'odds': 3}
        '''
        if self._index >= self._max_index:  # avoid recreating actions
            result = self._recent_possible_actions
        else:
            actions = itertools.product(
                *[component.generate(self._index)
                  for component in self._components.values()])
            self._index += 1
            result = self._recent_possible_actions = tuple(FeatureArray(a)
                                                           for a in actions)

        return result

    @property
    def components(self):
        return self._components.keys()

    @property
    def lower(self) -> Dict[str, Numerical]:
        return {component.name: component.lower  # type: ignore
                for component in self._components.values()
                if hasattr(component, 'lower')}

    @property
    def upper(self) -> Dict[str, Numerical]:
        return {component.name: component.upper  # type: ignore
                for component in self._components.values()
                if hasattr(component, 'upper')}

    @property
    def categories(self) -> Dict[str, Tuple[Categorical, ...]]:
        return {component.name: component.categories  # type: ignore
                for component in self._components.values()
                if hasattr(component, 'categories')}

    def reset(self) -> None:
        ''' Resets the generator.'''
        self._index: int = 0
        self._recent_possible_actions: Tuple[FeatureArray, ...] = ()


if __name__ == "__main__":
    AG = ActionGenerator[str, int]()
    AG.add_categorical(
        component_name='compass_directions',
        possible_values=(('N',), ('E', 'W'), ('N', 'S')),
        categories=('N', 'S', 'E', 'W'))
    AG.add_numerical(
        component_name='odds',
        possible_values=((1,), (1, 3)),  # type: ignore
        lower=1, upper=9)
    for i in ('1st', '2nd', '3rd', '4th'):
        print(f'calling possible_actions for the {i} time:')
        for action in AG.possible_actions():
            print(action.value)



# -*- coding: utf-8 -*-
'''
Aurora class
============

Aurora Dosing Protocol, based on `Ravvaz et al. (2017)
<https://doi.org/10.1161/circgenetics.117.001804>`_
'''

from typing import Any, Dict, List, Tuple

import reil.healthcare.dosing_protocols.dosing_protocol as dp


class Aurora(dp.DosingProtocol):
    '''
    Aurora Dosing Protocol, based on `Ravvaz et al. (2017)
    <https://doi.org/10.1161/circgenetics.117.001804>`_
    '''

    def prescribe(self,  # noqa: C901
                  patient: Dict[str, Any],
                  additional_info: dp.AdditionalInfo
                  ) -> Tuple[dp.DosingDecision, dp.AdditionalInfo]:
        '''
        Prescribe a dose for the given `patient` and `additional_info`.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics including:
            - age
            - day
            - INR_history
            - dose_history
            - interval_history

        additional_info:
            A dictionary of information being communicated between protocols at
            each call to `prescribe`. These additional information are
            protocol-dependent.

        Returns
        -------
        :
            A `DosingDecision` along with updated `additional_info`.
        '''
        today: int = patient['day']
        INRs: List[float] = patient['INR_history']
        previous_INR = INRs[-1]

        # to suppress Pylance's unbound variable error.
        previous_dose = 0.0
        previous_interval: int = 1
        next_interval: int

        if today > 1:
            previous_dose: float = patient['dose_history'][-1]
            previous_interval: int = patient['interval_history'][-1]

        red_flag: bool = additional_info.get('red_flag', False)
        skip_dose: int = additional_info.get('skip_dose', 0)
        new_dose: float = additional_info.get('new_dose', 0.0)
        number_of_stable_days: int = additional_info.get(
            'number_of_stable_days', 0)

        if red_flag:
            if previous_INR > 3.0:
                next_dose = 0.0
                next_interval = 2
            else:
                red_flag = False
                next_dose = new_dose
                next_interval = 7
        elif skip_dose:
            skip_dose = 0
            next_dose = new_dose
            next_interval = 7
        elif today <= 2:  # initial dosing
            # Note: in the actual described protocol, INR>1.1 and some other
            # conditions should get 5.0, but in their simulation, Ravvaz et al,
            # assumed all <65 get 10.0. (Footnote on page 10 of paper's
            # appendix)
            next_dose = 10.0 if patient['age'] < 65.0 else 5.0
            next_interval = 3 - today
        elif today == 3:  # adjustment dosing
            # Note: the paper is a bit vague on interval, since it has provided
            # the whole dosing table for days 3 and 4 (which says retest in 7
            # and 28 days), but at the same time says dose for days 3 and 4 are
            # based on the table. Here I assume that the interval is not
            # determined by the table, and only the dose is adjusted.
            next_interval = 2

            if previous_INR >= 2.0:
                next_dose = 5.0
                if previous_INR <= 3.0:
                    if 2.0 <= INRs[-2] <= 3.0:
                        number_of_stable_days = previous_interval
                    else:
                        number_of_stable_days = 1
                    # number_of_stable_days = self._stable_days(
                    #     INRs[-2], previous_INR, previous_interval)
            else:
                next_dose, _, _, _ = self.aurora_dosing_table(
                    previous_INR, previous_dose)
        elif today == 4:
            raise ValueError(
                'Cannot use Aurora on day 4. '
                'Dose on day 4 equals dose on day 3.')
        else:  # maintenance dosing
            if 2.0 <= previous_INR <= 3.0:
                if (previous_dose == patient['dose_history'][-2]) and (
                        2.0 <= INRs[-2] <= 3.0):
                    # stable dose and therapeutic INR
                    number_of_stable_days += previous_interval
                    # number_of_stable_days += self._stable_days(
                    #     INRs[-2], previous_INR, previous_interval)
                else:
                    number_of_stable_days = 1

                next_dose = previous_dose
                next_interval = self.aurora_retesting_table(
                    number_of_stable_days)
            else:
                number_of_stable_days = 0
                new_dose, next_interval, skip_dose, red_flag = \
                    self.aurora_dosing_table(previous_INR, previous_dose)
                if red_flag:
                    next_dose = 0.0
                    next_interval = 2
                elif skip_dose:
                    next_dose = 0.0
                    next_interval = skip_dose
                else:
                    next_dose = new_dose

        additional_info.update({
            'red_flag': red_flag,
            'skip_dose': skip_dose,
            'new_dose': new_dose,
            'number_of_stable_days': number_of_stable_days})

        return dp.DosingDecision(next_dose, next_interval), additional_info

    @staticmethod
    def _stable_days(INR_start: float, INR_end: float, interval: int) -> int:
        '''
        Compute the number of stable days.

        Arguments
        ---------
        INR_start:
            INR at the beginning of the period.

        INR_end:
            INR at the end of the period.

        interval:
            The number of days from start to end.

        Returns
        -------
        :
            The number of days in therapeuric range (TTR).

        Notes
        -----
        The paper is a bit vague! Initially I assumed we interpolate INRs to
        get the number of stable days, but it does not seem to be the case.
        So, in the updated implementation, if INR_end is not in the range,
        it returns 0, if it is in the range, but INR_start is not in the range,
        it returns 1, otherwise, it returns `interval`.

        '''
        # return sum(2 <= INR_end + (INR_start - INR_end) * i / interval <= 3
        #            for i in range(1, interval+1))
        if 2.0 <= INR_end <= 3.0:
            if 2.0 <= INR_start <= 3.0:
                return interval
            else:
                return 1

        return 0

    @staticmethod
    def aurora_dosing_table(
            current_INR: float, dose: float) -> Tuple[float, int, int, bool]:
        '''
        Determine the dosing information, based on Aurora dosing table.

        Arguments
        ---------
        current_INR:
            The latest value of INR.

        dose:
            The latest dose prescribed.

        Returns
        -------
        :
            * The next dose
            * The time of the next test (in days).
            * The number of doses to skip.
            * Red flag for too high INR values.
        '''
        skip_dose: int = 0
        red_flag = False
        _current_INR = round(current_INR, 2)
        next_test: int = 7

        if _current_INR < 1.50:
            dose = dose * 1.15
        elif _current_INR < 1.80:
            dose = dose * 1.10
        elif _current_INR < 2.00:
            dose = dose * 1.075
        elif _current_INR <= 3.00:
            next_test = 28
        elif _current_INR < 3.40:
            dose = dose * 0.925
        elif _current_INR < 4.00:
            dose = dose * 0.9
        elif _current_INR <= 5.00:
            skip_dose = 2
            dose = dose * 0.875
        else:
            red_flag = True
            next_test = 2
            dose = dose * 0.85

        return dose, next_test, skip_dose, red_flag

    @staticmethod
    def aurora_retesting_table(number_of_stable_days: int) -> int:
        '''
        Determine when the next test should be based on current number of
        stable days.

        Arguments
        ---------
        number_of_stable_days:
            Number of consecutive stable days.

        Returns
        -------
        :
            The time of the next test (in days).
        '''
        retesting_table = {1: 1, 2: 5, 7: 7, 14: 14, 28: 28}
        return retesting_table[max(
            map(lambda x: x if x <= number_of_stable_days else 1,
                retesting_table))]



# -*- coding: utf-8 -*-
'''
FrozenLake class
=================

This class creates a frozen lake (board) in which one square is the goal.
The `agent` starts from a location and should find the fastest route to
the goal. Some locations are holes.
'''

from random import choice
from typing import Any, Dict, List, Optional, Tuple

from reil.utils.mnkboard import MNKBoard
from reil.subjects.subject import Subject
from reil.datatypes import ReilData


class FrozenLake(MNKBoard, Subject):
    '''
    Build an m-by-n grid (using mnkboard super class) in which 1 player can play.
    Player wins if it can get to the goal square. Each element in the graph can be:
        S: starting point, safe
        F: frozen surface, safe
        H: hole (end with reward -1)
        G: goal (end with reward 1)

    Attributes
    ----------
        is_terminated: whether the game finished or not.
        possible_actions: a list of possible actions.

    Methods
    -------
        register: register the player and return its ID or return ID of current player.
        take_effect: moves the player on the grid.
        reset: clears the grid.
    '''
    def __init__(self, map: Optional[List[List[str]]] = None, **kwargs: Any):
        '''
        Arguments
        ---------
        map:
            the map to be used
        '''

        default_map = [['S', 'F', 'F', 'F'],
                       ['F', 'H', 'F', 'H'],
                       ['F', 'F', 'F', 'H'],
                       ['H', 'F', 'F', 'G']]

        def locate(map: List[List[Any]], element: Any) -> Tuple[int, int]:
            row = [element in m_i
                   for m_i in map].index(True)
            col = map[row].index(element)
            return (row, col)

        self._map = map if map is not None else default_map

        self._dim = (len(self._map), len(self._map[0]))
        self._start = locate(self._map, 'S')
        self._goal = locate(self._map, 'G')

        moves = ('U', 'D', 'R', 'L')
        self._default_moves = tuple(
            ReilData.single_categorical(
                name='move', value=m, categories=moves)
            for m in moves)

        MNKBoard.__init__(self, m=self._dim[0], n=self._dim[1], players=1)
        Subject.__init__(**kwargs)
        self.reset()

    def is_terminated(self, _id: Optional[int] = None) -> bool:
        '''Return True if the player get to the goal.'''
        return self._player_location == self._goal

    def possible_actions(self, _id: int = 0) -> Tuple[ReilData, ...]:
        '''Return the set of possible moves.'''
        return self._default_moves

    def default_reward(self, _id: int = 0) -> ReilData:
        return ReilData.single_base(name='reward', value=(float(self._player_location == self._goal) - 0.5) * 2)

    def register(self, player_name):
        '''
        Register an agent and return its ID.

        If the agent is new, a new ID is generated and the agent_name is added to agent_list.
        Arguments
        ---------
            agent_name: the name of the agent to be registered.

        Raises ValueError for an attempt to register more than one agent.
        '''
        if len(self._agent_list) == 0:
            return Subject.register(self, player_name)
        raise ValueError('Windy Gridworld only accepts one player.')

    def take_effect(self, action: ReilData, _id: int = 0) -> None:
        '''
        Move according to the action.

        Arguments
        ---------
            _id: ID of the player.
            action: the location in which the piece is set. Can be either in index format or row column format.
        '''
        Subject.take_effect(self, action, _id)

        row, column = self._player_location
        max_row = self._dim[0] - 1
        max_column = self._dim[1] - 1
        MNKBoard.clear_square(self, row=row, column=column)

        a = str(*action.value)
        temp = (
            row - (a == ['U', 'UR', 'UL']) + (a in ['D', 'DR', 'DL']),
            column - (a in ['L', 'UL', 'DL']) + (a in ['R', 'UR', 'DR'])
        )

        self._player_location = (
            min(max(temp[0], 0), max_row),
            min(max(temp[1], 0), max_column)
        )

        if self._map[self._player_location[0]][self._player_location[1]] == 'H':
            self._player_location = self._start

        MNKBoard.set_piece(self, player=1,
                           row=self._player_location[0],
                           column=self._player_location[1])

    def reset(self):
        '''Clear the board and update board_status.'''
        Subject.reset(self)
        MNKBoard.reset(self)

        self._player_location = self._start
        MNKBoard.set_piece(self, player=1,
                           row=self._player_location[0],
                           column=self._player_location[1])

    def _generate_state_components(self) -> None:
        def full_map(**kwargs: Any) -> Dict[str, Any]:
            return {'name': 'full_map',
                    'value': self._map,
                    'categories': ('S', 'F', 'H', 'G')}

        self._available_state_components = {
            'full_map': full_map,
        }

if __name__ == '__main__':
    board = FrozenLake()
    _ = board.register('P1')
    for _ in range(10):
        print(board.state)
        my_action = choice(board.possible_actions)
        board.take_effect(my_action, 1)
        print(my_action.value)
        print(f'{board}')



# -*- coding: utf-8 -*-
'''
Sink class
==========

A dummy buffer that does nothing!
'''

from typing import Dict, List, Optional, Tuple, Union

from reil.datatypes.buffers.buffer import Buffer, T1, T2, PickModes


class Sink(Buffer[T1, T2]):
    '''
    A sink class.
    '''
    def __init__(self, buffer_names: Optional[List[str]] = None) -> None:
        '''
        Arguments
        ---------
        buffer_names:
            A list containing the names of buffer queues.
        '''
        self._buffer_names = None
        self.setup(buffer_names=buffer_names)

    def setup(
            self, buffer_size: Optional[int] = None,
            buffer_names: Optional[List[str]] = None,
            pick_mode: Optional[PickModes] = None,
            clear_buffer: Optional[bool] = None) -> None:
        '''
        Set up the buffer.

        Arguments
        ---------
        buffer_size:
            The size of the buffer. This argument is only available for
            signature consistency. Assigning it has no effect.

        buffer_names:
            A list containing the names of buffer elements.

        pick_mode:
            The default mode to pick items from the list. This argument is
            only available for signature consistency. Assigning it has
            no effect.

        clear_buffer:
            Whether to clear the buffer when `reset` is called. This argument
            is only available for signature consistency. Assigning it has no
            effect.

        Raises
        ------
        ValueError:
            Cannot modify `buffer_size`. The value is already set.

        ValueError:
            Cannot modify `buffer_names`. The value is already set.

        Notes
        -----
        `setup` should be used only for attributes of the buffer that are
        not defined. Attempt to use `setup` to modify size, names or mode will
        result in an exception.
        '''
        super().setup(buffer_names=buffer_names)

    def add(self, data: Dict[str, Union[T1, T2]]) -> None:
        '''
        Append a new item to the buffer.

        Arguments
        ---------
        data:
            A dictionary with the name of buffer queues as keys.

        Notes
        -----
        This implementation of `add` does not check if the buffer is full
        or if the provided names exist in the buffer queues. As a result, this
        situations will result in exceptions by the system.
        '''
        return

    def pick(
            self,
            count: Optional[int] = None,
            mode: Optional[PickModes] = None
    ) -> Dict[str, Union[Tuple[T1, ...], Tuple[T2, ...]]]:
        '''
        Raises an exception.

        Arguments
        ---------
        count:
            The number of items to return. If omitted, the number of items in
            the buffer is used. `count` will be ignored if `mode` is 'all'.

        mode:
            How to pick items. If omitted, the default `pick_mode` specified
            during initialization or setup is used.

        Raises
        ------
        TypeError:
            Cannot pick from a `Sink`.
        '''
        raise TypeError('Cannot pick from a Sink.')

    def reset(self) -> None:
        '''
        Reset the buffer.
        '''
        pass



# -*- coding: utf-8 -*-
'''
DosingProtocol class
====================

The base class for all basic dosing protocols.
'''

import dataclasses
from typing import Any, Dict, Optional, Tuple


@dataclasses.dataclass
class DosingDecision:
    dose: float
    duration: Optional[int] = None


AdditionalInfo = Dict[str, Any]


class DosingProtocol:
    '''
    Base class for all dosing protocol objects.
    '''

    def __init__(self) -> None:
        pass

    def prescribe(self,
                  patient: Dict[str, Any],
                  additional_info: AdditionalInfo
                  ) -> Tuple[DosingDecision, AdditionalInfo]:
        '''
        Prescribe a dose for the given `patient` and `additional_info`.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics necessary to make dosing
            decisions.

        additional_info:
            A dictionary of information being communicated between protocols at
            each call to `prescribe`. These additional information are
            protocol-dependent.

        Returns
        -------
        :
            A `DosingDecision` along with updated `additional_info`.
        '''
        raise NotImplementedError

    def reset(self) -> None:
        '''Reset the dosing protocol'''
        pass



# -*- coding: utf-8 -*-
'''
Environment class
==================

The base class of all learning environments in which one or more `agents` act
on one or more `subjects`.
'''
import inspect
import pathlib
from collections import defaultdict
from typing import Any, Dict, Generator, List, Optional, Tuple, TypeVar, Union

import pandas as pd
from reil import stateful
from reil.agents.agent import Agent
from reil.agents.agent_demon import AgentDemon
from reil.datatypes.feature import FeatureArray
from reil.datatypes.interaction_protocol import InteractionProtocol
from reil.subjects.subject import Subject
from reil.subjects.subject_demon import SubjectDemon
from reil.utils.instance_generator import InstanceGenerator

T = TypeVar('T')

AgentSubjectTuple = Tuple[str, str]
EntityType = Union[Agent[T], Subject]
EntityGenType = Union[InstanceGenerator[Agent[T]],
                      InstanceGenerator[AgentDemon[T]],
                      InstanceGenerator[Subject],
                      InstanceGenerator[SubjectDemon]]


class Environment(stateful.Stateful):
    '''
    Provide an interaction and learning environment for `agents` and
    `subjects`.

    Notes
    -----
    `Agents` act on `subjects` and receive the reward of their action and
    the new state of those `subjects`. Then `agents` learn based on this
    information to improve their actions.

    `Pass`: visiting all protocols in the interaction sequence, once.

    `iteration`: For each subject, an iteration is one time reaching its
    terminal state. If the subject is an instance generator, then
    the generator should reach to terminal state, not just its
    current instance.
    '''

    def __init__(
            self,
            entity_dict: Optional[Dict[str, Union[
                EntityType[Any], EntityGenType[Any], str]]] = None,
            demon_dict: Optional[Dict[str, Union[
                AgentDemon[Any], SubjectDemon, str]]] = None,
            **kwargs: Any):
        '''
        Arguments
        ---------
        entity_dict:
            a dictionary that contains `agents`, `subjects`, and
            `generators`.
        '''
        super().__init__(**kwargs)

        self._agents: Dict[str, Agent[Any]] = {}
        self._subjects: Dict[str, Subject] = {}
        self._agent_demons: Dict[str, AgentDemon[Any]] = {}
        self._subject_demons: Dict[str, SubjectDemon] = {}
        self._instance_generators: Dict[str, EntityGenType[Any]] = {}
        self._assignment_list: Dict[
            AgentSubjectTuple,
            Tuple[Union[int, None], Union[int, None]]] = \
            defaultdict(lambda: (None, None))
        self._iterations: Dict[str, int] = defaultdict(int)
        self._agent_observers: Dict[
            Tuple[str, str],
            Generator[Union[FeatureArray, None], Any, None]] = {}

        if entity_dict is not None:
            self.add_entities(entity_dict)

        if demon_dict is not None:
            self.add_demons(demon_dict)

    def add_entities(
            self,
            entity_dict: Dict[str, Union[
                EntityType[Any], EntityGenType[Any], str]]) -> None:
        '''
        Add `agents` and `subjects` to the environment.

        Arguments
        ---------
        entity_dict:
            a dictionary consist of `agent`/ `subject` name and the
            respective entity. Names should be unique, otherwise overwritten.
            To assign one entity to multiple names, use the name in the
            first assignment as the value of the dict for other assignments.
            For example:
            >>> env.add_entities({'agent_1': Agent(), 'agent_2': 'agent_1'})

            When using name as value, the name is being looked up first in
            instance generators, then agents, and finally subjects. Whichever
            contains the name first, the entity corresponding to that instance
            is being used.

        Notes
        -----
        Reusing an `InstanceGenerator` produces unintended consequences.

        Raises
        ------
        ValueError
            An entity is being reused without being defined first.

        TypeError
            The entity is niether an `agent` [generator] nor a `subject`
            [generator].
        '''
        for name, obj in entity_dict.items():
            if isinstance(obj, str):
                _obj = self._instance_generators.get(
                    obj, self._agents.get(
                        obj, self._subjects.get(obj)))
                if _obj is None:
                    raise ValueError(f'entity {obj} defined for {name} is '
                                     'not in the list of agents, subjects, '
                                     'and generators.')
            else:
                _obj = obj
            if isinstance(_obj, InstanceGenerator):
                self._instance_generators.update({name: _obj})

                _, instance = next(_obj)
                if isinstance(instance, Agent):
                    self._agents.update({name: instance})
                elif isinstance(instance, Subject):
                    self._subjects.update({name: instance})
                else:
                    raise TypeError(
                        f'entity {name} is niether an agent generator nor a '
                        'subject generator.')

            elif isinstance(_obj, Agent):
                self._agents.update({name: _obj})
            elif isinstance(_obj, Subject):
                self._subjects.update({name: _obj})
            else:
                raise TypeError(
                    f'entity {name} is niether an agent nor a subject.')

    def remove_entity(self, entity_names: Tuple[str, ...]) -> None:
        '''
        Remove `agents`, `subjects`, or `instance_generators` from
        the environment.

        Arguments
        ---------
        entity_names:
            A list of `agent`/ `subject` names to be deleted.

        Notes
        -----
        This method removes the item from both `agents` and `subjects`
        lists. Hence, it is not recommended to use the same name for both
        an `agent` and a `subject`.
        '''
        for name in entity_names:
            if name in self._agents:
                del self._agents[name]
            if name in self._subjects:
                del self._subjects[name]
            if name in self._instance_generators:
                del self._instance_generators[name]

    def add_demons(
            self,
            demon_dict: Dict[str, Union[
                AgentDemon[Any], SubjectDemon, str]]) -> None:
        '''
        Add `AgentDemon`s and `SubjectDemon`s to the environment.

        Arguments
        ---------
        demon_dict:
            a dictionary consist of `AgentDemon`/ `SubjectDemon` name and the
            respective entity. Names should be unique, otherwise overwritten.
            To assign one entity to multiple names, use the name in the
            first assignment as the value of the dict for other assignments.
            For example:
            >>> env.add_demons({'agent_demon_1': AgentDemon(),
                                'agent_demon_2': 'agent_demon_1'})

            When using name as value, the name is being looked up first in
            agent demons, and then subject demons. Whichever
            contains the name first, the demon corresponding to that instance
            is being used.

        Raises
        ------
        ValueError
            A demon is being reused without being defined first.

        TypeError
            The demon is niether an `AgentDemon` nor a `SubjectDemon`.
        '''
        for name, obj in demon_dict.items():
            if isinstance(obj, str):
                _obj = self._agent_demons.get(
                        obj, self._subject_demons.get(obj))
                if _obj is None:
                    raise ValueError(f'entity {obj} defined for {name} is '
                                     'not in the list of agent demons, '
                                     'and subject demons.')
            else:
                _obj = obj
            if isinstance(_obj, AgentDemon):
                self._agent_demons.update({name: _obj})
            elif isinstance(_obj, SubjectDemon):
                self._subject_demons.update({name: _obj})
            else:
                raise TypeError(
                    f'entity {name} is niether an agent demon nor '
                    'a subject demon.')

    def remove_demon(self, demon_names: Tuple[str, ...]) -> None:
        '''
        Remove `agent demons` or `subject demons` from
        the environment.

        Arguments
        ---------
        demon_names:
            A list of demon names to be deleted.

        Notes
        -----
        This method removes the item from both `agent_demons` and
        `subject_demons` lists.
        Hence, it is not recommended to use the same name for both
        an `agent demon` and a `subject demon`.
        '''
        for name in demon_names:
            if name in self._agent_demons:
                del self._agent_demons[name]
            if name in self._subject_demons:
                del self._subject_demons[name]

    def simulate_pass(self, n: int = 1) -> None:
        '''
        Go through the interaction sequence for a number of passes and
        simulate interactions accordingly.

        Arguments
        ---------
        n:
            The number of passes that simulation should go.
        '''
        raise NotImplementedError

    def simulate_to_termination(self) -> None:
        '''
        Go through the interaction sequence and simulate interactions
        accordingly, until all `subjects` are terminated.

        Notes
        -----
        To avoid possible infinite loops caused by normal `subjects`,
        this method is only available if all `subjects` are generated
        by `instance generators`.

        Raises
        ------
        TypeError:
            Attempt to call this method will normal subjects in the interaction
            sequence.
        '''
        raise NotImplementedError

    @classmethod
    def interact(
            cls,
            agent_id: int,
            agent_observer: Generator[Union[FeatureArray, None], Any, None],
            subject_instance: Union[Subject, SubjectDemon],
            state_name: str,
            action_name: str,
            reward_name: str,
            iteration: int,
            times: int = 1) -> None:
        '''
        Allow `agent` and `subject` to interact at most `times` times.

        Attributes
        ----------
        agent_id:
            Agent's ID by which it is registered at the subject.

        subject_id:
            Subject's ID by which it is registered at the `agent`.

        agent_instance:
            An instance of an `agent` that takes the action.

        subject_instance:
            An instance of a `subject` that computes reward, determines
            possible actions, and takes the action.

        state_name:
            A string that specifies the state definition.

        action_name:
            A string that specifies the action definition.

        reward_name:
            A string that specifies the reward function definition.

        iteration:
            The iteration of of the current run. This value is used by the
            `agent` to determine the action.

        times:
            The number of times the `agent` and the `subject` should interact.

        Returns
        -------
        :
            A list of subject's reward and state before taking an action
            and agent's action.

        Notes
        -----
        If subject is terminated before "times" iterations, the result will
        be truncated and returned. In other words, the output will not
        necessarily have a lenght of "times".
        '''
        for _ in range(times):
            reward = subject_instance.reward(
                name=reward_name, _id=agent_id)
            agent_observer.send(reward)

            state = subject_instance.state(name=state_name, _id=agent_id)
            possible_actions = subject_instance.possible_actions(
                name=action_name, _id=agent_id)
            if possible_actions:
                action = agent_observer.send({'state': state,
                                              'actions': possible_actions,
                                              'iteration': iteration})
                subject_instance.take_effect(action, agent_id)  # type: ignore

    @classmethod
    def interact_while(
            cls,
            agent_id: int,
            agent_observer: Generator[Union[FeatureArray, None], Any, None],
            subject_instance: Union[Subject, SubjectDemon],
            state_name: str,
            action_name: str,
            reward_name: str,
            iteration: int) -> None:
        '''
        Allow `agent` and `subject` to interact until `subject` is terminated.

        Attributes
        ----------
        agent_id:
            Agent's ID by which it is registered at the subject.

        agent_instance:
            An instance of an agent that takes the action.

        subject_instance:
            An instance of a subject that computes reward,
            determines possible actions, and takes the action.

        state_name:
            A string that specifies the state definition.

        action_name:
            A string that specifies the action definition.

        reward_name:
            A string that specifies the reward function definition.

        iteration:
            The iteration of of the current run. This value is used by the
            agent to determine the action.

        Returns
        -------
        :
            A list of subject's reward and state before taking an action and
            agent's action.

        Notes
        -----
        For `instance generators`, only the current
        instance is run to termination, not the whole generator.
        '''
        while not subject_instance.is_terminated(agent_id):
            cls.interact(agent_id, agent_observer, subject_instance,
                         state_name, action_name, reward_name,
                         iteration)

    def assert_protocol(self, protocol: InteractionProtocol) -> None:
        '''
        Check whether the given protocol:

        * contains only entities that are known to the `environment`.

        * unit is one of the possible values.

        Arguments
        ---------
        protocol:
            An interaction protocol.

        Raises
        ------
        ValueError
            `agent`, `agent demon`, `subject`, or `subject demon` is not
            defined.

        ValueError
            `unit` is not one of `interaction`, `instance`, or `iteration`.
        '''
        agent_name = protocol.agent.name
        agent_demon_name = protocol.agent.demon_name
        subject_name = protocol.subject.name
        subject_demon_name = protocol.subject.demon_name
        unit = protocol.unit

        if agent_name not in self._agents:
            raise ValueError(f'Unknown agent name: {agent_name}.')
        if (agent_demon_name is not None
                and agent_demon_name not in self._agent_demons):
            raise ValueError(
                f'Unknown agent demon name: {agent_demon_name}.')
        if subject_name not in self._subjects:
            raise ValueError(f'Unknown subject name: {subject_name}.')
        if (subject_demon_name is not None
                and subject_demon_name not in self._subject_demons):
            raise ValueError(
                f'Unknown subject demon name: {subject_demon_name}.')
        if unit not in ('interaction', 'instance', 'iteration'):
            raise ValueError(
                f'Unknown unit: {unit}. '
                'It should be one of interaction, instance, or iteration. '
                'For subjects of non-instance generator, iteration and '
                'instance are equivalent.')

    def register(self,
                 interaction_protocol: InteractionProtocol,
                 get_agent_observer: bool = False) -> None:
        '''
        Register the `agent` and `subject` of an interaction protocol.

        Arguments
        ---------
        interaction_protocol:
            The protocol whose `agent` and `subject` should be registered.

        get_agent_observer:
            If `True`, the method calls the `observe` method of the `agent`
            with `subject_id`, and adds the resulting generator to the list
            of observers.

        Notes
        -----
        When registration happens for the first time, agents and subjects
        get any ID that the counterpart provides. However, in the follow up
        registrations, `entities` attempt to register with the same ID to
        have access to the same information.
        '''
        a_name = interaction_protocol.agent.name
        a_demon_name = interaction_protocol.agent.demon_name
        a_stat = interaction_protocol.agent.statistic_name
        s_name = interaction_protocol.subject.name
        s_demon_name = interaction_protocol.subject.demon_name
        a_s_name = (a_name, s_name)

        if s_demon_name is None:
            subject_instance = self._subjects[s_name]
        else:
            subject_instance = \
                self._subject_demons[s_demon_name](
                    self._subjects[s_name])

        if a_demon_name is None:
            agent_instance = self._agents[a_name]
        else:
            agent_instance = \
                self._agent_demons[a_demon_name](
                    self._agents[a_name])

        a_id, s_id = self._assignment_list[a_s_name]
        a_id = subject_instance.register(entity_name=a_name, _id=a_id)
        s_id = agent_instance.register(entity_name=s_name, _id=s_id)

        self._assignment_list[a_s_name] = (a_id, s_id)

        if get_agent_observer:
            self._agent_observers[a_s_name] = \
                agent_instance.observe(s_id, a_stat)

    def close_agent_observer(self, protocol: InteractionProtocol) -> None:
        '''
        Close an `agent_observer` corresponding to `protocol`.

        Before closing the observer, the final `reward` and `state` of the
        system are passed on to the observer.

        Attributes
        -----------
        protocol:
            The protocol whose `agent_observer` should be closed.

        Notes
        -----
        This method should only be used if a `subject` is terminated.
        Otherwise, the `agent_observer` might be expecting to receive different
        values, and it will corrupt the training data for the `agent`.
        '''
        agent_name = protocol.agent.name
        subject_name = protocol.subject.name
        r_func_name = protocol.reward_name
        state_name = protocol.state_name
        a_s_names = (agent_name, subject_name)
        s_demon_name = protocol.subject.demon_name

        if s_demon_name is None:
            subject_instance = self._subjects[subject_name]
        else:
            subject_instance = \
                self._subject_demons[s_demon_name](
                    self._subjects[subject_name])

        if inspect.getgeneratorstate(
                self._agent_observers[a_s_names]) != inspect.GEN_SUSPENDED:
            return

        a_id, _ = self._assignment_list[a_s_names]
        reward = subject_instance.reward(
            name=r_func_name, _id=a_id)
        state = subject_instance.state(
            name=state_name, _id=a_id)

        self._agent_observers[a_s_names].send(reward)
        self._agent_observers[a_s_names].send({'state': state,
                                               'actions': None,
                                               'iteration': None})
        self._agent_observers[a_s_names].close()

    def reset_subject(self, subject_name: str) -> bool:
        '''
        When a `subject` is terminated for all interacting `agents`, this
        function is called to reset the `subject`.

        If the `subject` is an `InstanceGenerator`, a new instance is created.
        If reset is successful, `iteration` is incremented by one.

        Attributes
        ----------
        subject_name:
            Name of the `subject` that is terminated.

        Returns
        -------
        :
            `True` if the `instance_generator` for the `subject` is still
            active, `False` if it hit `StopIteration`.

        Notes
        -----
        `Environment.reset_subject` only resets the `subject`. It does not
        get the statistics for that `subject`.
        '''
        if subject_name in self._instance_generators:
            # get a new instance if possible,
            # if not instance generator returns StopIteration.
            # So, increment iteration by 1, then if the generator is not
            # terminated, get a new instance.
            # If the generator is terminated, check if it is finite. If
            # infinite, call it again to get a subject. If not, disable reward
            # for the current subject, so that agent_observer does not raise
            # exception.
            try:
                _, self._subjects[subject_name] = next(  # type: ignore
                    self._instance_generators[subject_name])

            except StopIteration:
                self._iterations[subject_name] += 1
                if self._instance_generators[subject_name].is_terminated():
                    self._subjects[subject_name].reward.disable()
                else:
                    _, self._subjects[subject_name] = next(  # type: ignore
                        self._instance_generators[subject_name]
                        )
                return False
        else:
            self._iterations[subject_name] += 1
            self._subjects[subject_name].reset()

        return True

    def load(self,  # noqa: C901
             entity_name: Union[List[str], str] = 'all',
             filename: Optional[str] = None,
             path: Optional[Union[str, pathlib.PurePath]] = None) -> None:
        '''
        Load an entity or an `environment` from a file.

        Arguments
        ---------
        filename:
            The name of the file to be loaded.

        entity_name:
            If specified, that entity (`agent` or `subject`) is being
            loaded from file. 'all' loads an `environment`.

        Raises
        ------
        ValueError
            The filename is not specified.
        '''
        _filename: str = filename or self._name
        _path = pathlib.Path(path or self._path)

        if entity_name == 'all':
            super().load(filename=_filename, path=_path)
            self._instance_generators: Dict[str, EntityGenType[Any]] = {}
            self._agents: Dict[str, Agent[Any]] = {}
            self._subjects: Dict[str, Subject] = {}
            self._agent_demons: Dict[str, AgentDemon[Any]] = {}
            self._subject_demons: Dict[str, SubjectDemon] = {}

            for name, obj_type in self._env_data['instance_generators']:
                self._instance_generators[name] = obj_type.from_pickle(
                    path=(_path / f'{_filename}.instance_generators'),
                    filename=name)

            for name, obj_type in self._env_data['agents']:
                if name in self._instance_generators:
                    self._agents[name] = (  # type: ignore
                        self._instance_generators[name]._object)
                else:
                    self._agents[name] = obj_type.from_pickle(
                        path=(_path / f'{_filename}.agents'), filename=name)

            for name, obj_type in self._env_data['subjects']:
                if name in self._instance_generators:
                    self._subjects[name] = (  # type: ignore
                        self._instance_generators[name]._object)
                else:
                    self._subjects[name] = obj_type.from_pickle(
                        path=(_path / f'{_filename}.subjects'), filename=name)

            for name, obj_type in self._env_data['agent_demons']:
                self._agent_demons[name] = obj_type.from_pickle(
                    path=(_path / f'{_filename}.agent_demons'), filename=name)

            for name, obj_type in self._env_data['subject_demons']:
                self._subject_demons[name] = obj_type.from_pickle(
                    path=(_path / f'{_filename}.subject_demons'),
                    filename=name)

            del self._env_data

        else:
            for obj in entity_name:
                if obj in self._instance_generators:
                    self._instance_generators[obj].load(
                        path=(_path / f'{_filename}.instance_generators'),
                        filename=obj)
                    self._instance_generators[obj].reset()

                if obj in self._agents:
                    self._agents[obj].load(
                        path=(_path / f'{_filename}.agents'), filename=obj)
                    self._agents[obj].reset()
                elif obj in self._subjects:
                    self._subjects[obj].load(
                        path=(_path / f'{_filename}.subjects'), filename=obj)
                    self._subjects[obj].reset()
                elif obj in self._agent_demons:
                    self._agent_demons[obj].load(
                        path=(_path / f'{_filename}.agent_demons'),
                        filename=obj)
                    self._agent_demons[obj].reset()
                elif obj in self._subject_demons:
                    self._subject_demons[obj].load(
                        path=(_path / f'{_filename}.subject_demons'),
                        filename=obj)
                    self._subject_demons[obj].reset()

    def save(self,  # noqa: C901
             filename: Optional[str] = None,
             path: Optional[Union[str, pathlib.PurePath]] = None,
             data_to_save: Union[List[str], str] = 'all'
             ) -> Tuple[pathlib.PurePath, str]:
        '''
        Save an entity or the `environment` to a file.

        Arguments
        ---------
        filename:
            The name of the file to be saved.

        path:
            The path of the file to be saved.

        entity_name:
            If specified, that entity (`agent` or `subject`) is being saved
            to file. 'all' saves the `environment`.

        Raises
        ------
        ValueError
            The filename is not specified.
        '''
        _filename = filename or self._name
        _path = pathlib.Path(path or self._path)

        if data_to_save == 'all':
            open_observers = set(a_s_names
                                 for a_s_names in self._agent_observers
                                 if inspect.getgeneratorstate(
                                     self._agent_observers[a_s_names]
                                 ) not in [inspect.GEN_CREATED or
                                           inspect.GEN_CLOSED])
            if open_observers:
                raise RuntimeError('Cannot save an environment in '
                                   'the middle of a simulation. '
                                   'These agent/subject interactions '
                                   'are still underway:\n'
                                   f'{open_observers}')

            temp, self._agent_observers = (  # type: ignore
                self._agent_observers, None)

            self._env_data: Dict[str, List[Any]] = defaultdict(list)

            try:
                for name, entity in self._instance_generators.items():
                    _, filename = entity.save(
                        path=_path / f'{_filename}.instance_generators',
                        filename=name)
                    self._env_data['instance_generators'].append(
                        (filename, type(entity)))

                for name, agent in self._agents.items():
                    if name in self._instance_generators:
                        self._env_data['agents'].append((name, None))
                    else:
                        _, filename = agent.save(
                            path=_path / f'{_filename}.agents', filename=name)
                        self._env_data['agents'].append(
                            (filename, type(agent)))

                for name, subject in self._subjects.items():
                    if name in self._instance_generators:
                        self._env_data['subjects'].append((name, None))
                    else:
                        _, filename = subject.save(
                            path=_path / f'{_filename}.subjects',
                            filename=name)
                        self._env_data['subjects'].append(
                            (filename, type(subject)))

                for name, agent_demon in self._agent_demons.items():
                    _, filename = agent_demon.save(
                        path=_path / f'{_filename}.agent_demons',
                        filename=name)
                    self._env_data['agent_demons'].append(
                        (filename, type(agent_demon)))

                for name, subject_demon in self._subject_demons.items():
                    _, filename = subject_demon.save(
                        path=_path / f'{_filename}.subject_demons',
                        filename=name)
                    self._env_data['subject_demons'].append(
                        (filename, type(subject_demon)))

                super().save(
                    filename=_filename, path=_path,
                    data_to_save=tuple(v for v in self.__dict__
                                       if v not in ('_agents',
                                                    '_subjects',
                                                    '_instance_generators',
                                                    '_agent_demons',
                                                    '_subject_demons')))

                del self._env_data

            finally:
                self._agent_observers = temp

        else:
            for obj in data_to_save:
                if obj in self._instance_generators:
                    self._instance_generators[obj].save(
                        path=_path / f'{_filename}.instance_generators',
                        filename=obj)
                elif obj in self._agents:
                    self._agents[obj].save(
                        path=_path / f'{_filename}.agents', filename=obj)
                elif obj in self._subjects:
                    self._subjects[obj].save(
                        path=_path / f'{_filename}.subjects', filename=obj)
                elif obj in self._agent_demons:
                    self._agent_demons[obj].save(
                        path=_path / f'{_filename}.agent_demons',
                        filename=obj)
                elif obj in self._subject_demons:
                    self._subject_demons[obj].save(
                        path=_path / f'{_filename}.subject_demons',
                        filename=obj)
                else:
                    self._logger.warning(f'Cannot save {obj} individually. '
                                         'Try saving the whole environment.')

        return _path, _filename

    def report_statistics(self,
                          unstack: bool = True,
                          reset_history: bool = True
                          ) -> Dict[Tuple[str, str], pd.DataFrame]:
        '''Generate statistics for agents and subjects.

        Parameters
        ----------
        unstack:
            Whether to unstack the resulting pivottable or not.

        reset_history:
            Whether to clear up the history after computing stats.

        Returns
        -------
        :
            A dictionary with state-subject pairs as keys and dataframes as
            values.
        '''
        raise NotImplementedError

    def __repr__(self) -> str:
        try:
            return (
                super().__repr__() + '\n Agents:\n'
                '\n\t'.join((a.__repr__()
                             for a in self._agents.values())) +
                '\nSubjects:\n'
                '\n\t'.join((s.__repr__()
                             for s in self._subjects.values())) +
                '\n AgentDemons:\n'
                '\n\t'.join((a.__repr__()
                             for a in self._agent_demons.values())) +
                '\nSubjectDemons:\n'
                '\n\t'.join((s.__repr__()
                             for s in self._subject_demons.values()))
            )
        except AttributeError:
            return super().__repr__()



# -*- coding: utf-8 -*-
'''
legacy classes for reinforcement learning
=========================================

This module provides classes that are no longer supported.

Classes
-------
Environment:
    The base class that provides minimum required
    functionality for a reinforcement learning environment.

Experiment:
    An environment to explore performance of trained
    agents on subjects.

IterableSubject:
    Now `InstanceGenerator` does the job and goes beyond that!

LegacyReilData

Risk

Snake

WarfarinClusterAgent:
    an agent whose actions are based on clustering of observations.
    Note: This agent will no longer run due to changes in the module
    implementation.
'''



import argparse
import collections
import dataclasses
import pathlib
from typing import Any, Dict, List, Optional, Type, Union

from reil.reilbase import ReilBase
from ruamel.yaml import YAML


@dataclasses.dataclass
class CommandlineArgument:
    name: str
    type: Type[Any]
    default: Any


class CommandlineParser:
    def __init__(
            self,
            cmd_args: List[CommandlineArgument],
            extra_args: Optional[Dict[str, Any]] = None) -> None:

        self.parsed_args = vars(self._parse_cmd_args(cmd_args, extra_args))

    @staticmethod
    def _parse_cmd_args(cmd_args: List[CommandlineArgument],
                        extra_args: Optional[Dict[str, str]] = None
                        ) -> argparse.Namespace:
        '''
        Parse command line arguments, and add `extra_args`.
        '''
        arg_parser = argparse.ArgumentParser()

        for arg in cmd_args:
            temp = {'type': arg.type, 'default': arg.default}
            if isinstance(arg.default, (list, tuple)):
                temp['nargs'] = '+'
            arg_parser.add_argument(f'--{arg.name}', **temp)

        parsed_args = arg_parser.parse_args()

        if extra_args is not None:
            parsed_args = arg_parser.parse_args(
                list(b
                     for a in extra_args.items()
                     for b in a),
                namespace=parsed_args)

        return parsed_args


class ConfigParser:
    def __init__(
            self,
            config_filenames: Dict[str, str],
            config_path: Optional[Union[pathlib.Path, str]] = None,
            vars_dict: Optional[Dict[str, str]] = None) -> None:

        self.config: Dict[str, Any]
        if config_filenames:
            self.config = {
                key: self._load_config_file(value, config_path, vars_dict)
                for key, value in config_filenames.items()
            }
        else:
            self.config = collections.defaultdict()

    @staticmethod
    def _load_config_file(
            filename: str,
            path: Optional[Union[pathlib.Path, str]] = None,
            vars_dict: Optional[Dict[str, str]] = None) -> Any:

        _path = pathlib.Path(path or '.')
        _filename = filename if filename.endswith((
            '.yaml', '.yml')) else f'{filename}.yaml'

        with open(_path / _filename, 'r') as f:
            temp = f.read()
            # temp = YAML().load(f)

        if vars_dict:
            for name, value in vars_dict.items():
                temp = temp.replace(f'${name}$', str(value))

        return YAML().load(temp)  # type: ignore

    def extract(
            self, root_name: str, branch_name: str, as_object: bool = False
    ) -> Any:
        conf = self.config[root_name][branch_name]
        if as_object:
            return ReilBase.parse_yaml(conf)

        return conf

    def contains(self, root_name: str, branch_name: str) -> bool:
        return branch_name in self.config[root_name]



# -*- coding: utf-8 -*-
'''
Gage class
==========

Gage dosing protocol ('pharmacogenetic', 'clinical'), based on `
Gage et al. (2008)
<https://doi.org/10.1038/clpt.2008.10>`_
'''

from math import exp, sqrt
from typing import Any, Dict, Literal, Tuple

import reil.healthcare.dosing_protocols.dosing_protocol as dp


class Gage(dp.DosingProtocol):
    '''
    Gage dosing protocol ('pharmacogenetic', 'clinical'), based on `
    Gage et al. (2008)
    <https://doi.org/10.1038/clpt.2008.10>`_
    '''

    def __init__(self,
                 method: Literal['pharmacogenetic',
                                 'clinical'] = 'pharmacogenetic') -> None:
        '''
        Arguments
        ---------
        method:
            One of 'pharmacogenetic', 'clinical'.
        '''
        if method.lower() == 'clinical':
            self._method = self.clinical
        elif method.lower() in ['pg', 'pharmacogenetic', 'default']:
            self._method = self.pg
        else:
            raise ValueError(
                f'Unknown method {method}.\n'
                'Acceptable methods are clinical and pharmacogenetic.')

    def prescribe(self,  # noqa: C901
                  patient: Dict[str, Any],
                  additional_info: dp.AdditionalInfo
                  ) -> Tuple[dp.DosingDecision, dp.AdditionalInfo]:
        '''
        Prescribe a dose for the given `patient` and `additional_info`.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics including:
            - age
            - BSA (m^2)
            - CYP2C9 *
            - VKORC1 *
            - target_INR
            - race (Only 'Black' changes the dose.)
            - amiodarone ('Yes', 'No')
            - tobaco ('Yes', 'No')
            - DVT/PE ('Yes', 'No') (Deep Venous Thrombosis/ Pulmonary Embolism)

        additional_info:
            A dictionary of information being communicated between protocols at
            each call to `prescribe`.

        Returns
        -------
        :
            A `DosingDecision` along with updated `additional_info`.

        Notes
        -----
        * `CYP2C9` and `VKORC1` are only available in the
            pharmacogenetic model.
        '''
        dose = self._method(patient)

        return dp.DosingDecision(dose), additional_info

    @staticmethod
    def clinical(patient: Dict[str, Any]) -> float:
        '''
        Determine warfarin dose using clinical Gage formula.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics including:
            - age
            - BSA (m^2)
            - race (Only 'Black' changes the dose.)
            - amiodarone ('Yes', 'No')
            - tobaco ('Yes', 'No')
            - target_INR
            - DVT/PE ('Yes', 'No') (Deep Venous Thrombosis/ Pulmonary Embolism)

        Returns
        -------
        :
            The maintenance dose

        Notes
        -----
        If BSA is not provided, weight (lb) and height (in) is used to
        calculate BSA using Mosteller method.

        target_INR is assumed 2.5 if not provided.
        '''
        if (bsa := patient.get('BSA')) is None:
            try:
                bsa = sqrt(patient['height'] * 2.54 *
                           patient['weight'] * 0.454 / 3600)
            except KeyError:
                raise KeyError(
                    'Either `BSA` in m^2, or `weight` in lb and `height` in in'
                    'should be provided.')

        african_american = patient.get('race') == 'Black'
        target_INR = patient.get('target_INR', 2.5)
        amiodarone = patient.get('amiodarone') == 'Yes'
        smokes = patient.get('tobaco') == 'Yes'
        dvt_pe = patient.get('DVT/PE') == 'Yes'

        dose = exp(0.613
                   + 0.4250 * bsa
                   - 0.0075 * patient['age']
                   + 0.1560 * african_american
                   + 0.2160 * target_INR
                   - 0.2570 * amiodarone
                   + 0.1080 * smokes
                   + 0.0784 * dvt_pe)

        return dose

    @staticmethod
    def pg(patient: Dict[str, Any]) -> float:
        '''
        Determine warfarin dose using pharmacogenetic Gage formula.

        Arguments
        ---------
        patient:
            A dictionary of patient characteristics including:
            - age
            - BSA (m^2)
            - CYP2C9
            - VKORC1
            - target_INR
            - race (Only 'Black' changes the dose.)
            - amiodarone ('Yes', 'No')
            - tobaco ('Yes', 'No')
            - DVT/PE ('Yes', 'No') (Deep Venous Thrombosis/ Pulmonary Embolism)

        Returns
        -------
        :
            The maintenance dose

        Notes
        -----
        If BSA is not provided, weight (lb) and height (in) is used to
        calculate BSA using Mosteller method.

        target_INR is assumed 2.5 if not provided.
        '''
        if (bsa := patient.get('BSA')) is None:
            try:
                bsa = sqrt(patient['height'] * 2.54 *
                           patient['weight'] * 0.454 / 3600)
            except KeyError:
                raise KeyError(
                    'Either `BSA` in m^2, or `weight` in lb and `height` in in'
                    ' should be provided.')

        african_american = patient.get('race') == 'Black'
        target_INR = patient.get('target_INR', 2.5)
        amiodarone = patient.get('amiodarone') == 'Yes'
        smokes = patient.get('tobaco') == 'Yes'
        dvt_pe = patient.get('DVT/PE') == 'Yes'

        cyp2c9 = patient.get('CYP2C9', '')
        cyp2c9_2 = cyp2c9.count('*2')
        cyp2c9_3 = cyp2c9.count('*3')
        vkorc1_ga = patient['VKORC1'] == 'G/A'

        dose = exp(0.9751
                   - 0.3238 * vkorc1_ga
                   + 0.4317 * bsa
                   - 0.4008 * cyp2c9_3
                   - 0.00745 * patient['age']
                   - 0.2066 * cyp2c9_2
                   + 0.2029 * target_INR
                   - 0.2538 * amiodarone
                   + 0.0922 * smokes
                   - 0.0901 * african_american
                   + 0.0664 * dvt_pe)

        return dose



# -*- coding: utf-8 -*-
'''
exploration_strategies module
=============================

Contains classes that mimics exploration strategies in reinforcement learning.

Classes
-------
ExplorationStrategy:
    The base class for all exploration strategies.

ConstantEpsilonGreedy:
    An epsilon greedy object with constant epsilon

VariableEpsilonGreedy:
    An epsilon greedy object that accepts a uni-variate
    function to determine epsilon.
'''
import random
import warnings
from typing import Callable

from reil import reilbase


class ExplorationStrategy(reilbase.ReilBase):
    '''
    The base class for all exploration strategies.
    '''

    def __init__(self) -> None:
        pass

    def explore(self, iteration: int = 0) -> bool:
        '''
        Return `True` if the `agent` needs to explore.

        Arguments
        ---------
        iteration:
            The current iteration number.

        Returns
        -------
        :
            `True` if the caller should explore, otherwise `False`.
        '''
        return True


class ConstantEpsilonGreedy(ExplorationStrategy):
    '''
    An epsilon greedy object with constant epsilon.
    '''

    def __init__(self, epsilon: float = 0.0) -> None:
        '''
        Arguments
        ---------
        epsilon:
            The value of epsilon

        Notes
        -----
        If epsilon is not in the range of [0, 1], a warning is being issued,
        but it does not raise an exception.
        '''
        if not (0.0 <= epsilon <= 1.0):
            warnings.warn('epsilon is not in the range of [0, 1].')
        self._epsilon = epsilon

    def explore(self, iteration: int = 0) -> bool:
        '''
        Return `True` if a randomly generated number is less than `epsilon`.

        Arguments
        ---------
        iteration:
            The current iteration number.

        Returns
        -------
        :
            `True` if the caller should explore, otherwise `False`.
        '''
        return random.random() < self._epsilon


class VariableEpsilonGreedy(ExplorationStrategy):
    '''
    An epsilon greedy object with constant epsilon.
    '''

    def __init__(self, epsilon: Callable[[int], float]) -> None:
        '''
        Arguments
        ---------
        epsilon:
            A uni-variate function that computes `epsilon`
            based on `iteration`.

        Raises
        ------
        TypeError:
            `epsilon` is not callable.
        '''
        if not callable(epsilon):
            raise TypeError('epsilon should be callable. '
                            'For constant epsilon, '
                            'use `ConstantEpsilonGreedy` class.')
        self._epsilon = epsilon

    @classmethod
    def _empty_instance(cls):
        return cls(lambda e: 0.0)

    def explore(self, iteration: int) -> bool:
        '''
        Return `True` if a randomly generated number is less than `epsilon`.

        Arguments
        ---------
        iteration:
            The current iteration number.

        Returns
        -------
        :
            `True` if the caller should explore, otherwise `False`.
        '''
        return random.random() < self._epsilon(iteration)



# -*- coding: utf-8 -*-
'''
DeepQLearning class
===================

A Q-learning `agent` with a Neural Network Q-function approximator.
'''

from typing import Any

from reil.agents.q_learning import QLearning
from reil.datatypes.buffers import VanillaExperienceReplay
from reil.datatypes.feature import FeatureArray
from reil.learners import Dense
from reil.utils.exploration_strategies import ExplorationStrategy


class DeepQLearning(QLearning):
    '''
    A Deep Q-learning `agent`.
    '''

    def __init__(self,
                 learner: Dense,
                 buffer: VanillaExperienceReplay[FeatureArray, float],
                 exploration_strategy: ExplorationStrategy,
                 method: str = 'backward',
                 **kwargs: Any):
        '''
        Arguments
        ---------
        learner:
            The `Learner` of type `Dense` that does the learning.

        buffer:
            A buffer that collects observations for training. Some
            variation of `ExperienceReply` is recommended.

        exploration_strategy:
            An `ExplorationStrategy` object that determines
            whether the `action` should be exploratory or not for a given
            `state` at a given `iteration`.

        method:
            Either 'forward' or 'backward' Q-learning.

        kwargs:
            Keyword arguments to be passed on to the parent class.
        '''
        super().__init__(learner=learner,
                         buffer=buffer,
                         exploration_strategy=exploration_strategy,
                         method=method,
                         **kwargs)



import pathlib
import time
from typing import Dict, Optional, Tuple, Union

import pandas as pd


class OutputWriter:
    def __init__(self,
                 filename: str,
                 path: Union[str, pathlib.PurePath] = '.',
                 columns: Optional[Tuple[str]] = None
                 ) -> None:

        self._path = pathlib.PurePath(path)
        self._csv_filename = filename if filename.endswith((
            '.yaml', '.yml')) else f'{filename}.csv'
        pathlib.Path(self._path).mkdir(parents=True, exist_ok=True)
        if columns:
            with open(self._path / self._csv_filename, 'a+', newline='') as f:
                pd.DataFrame([], columns=columns).to_csv(
                    f, header=True)

    def write_stats_output(
            self, stats_output: Dict[Tuple[str, str], pd.DataFrame]) -> None:
        '''Write stats to file.'''
        attempts = 0
        while attempts < 5 and not self._write_stats_output(stats_output):
            time.sleep(1)
            attempts += 1

        if attempts == 5:
            with open(self._path / f'{self._csv_filename}_temp',
                      'a+', newline='') as f:
                for s in stats_output.values():
                    print(s)
                    s.to_csv(f, mode='a+', header=False)

    def _write_stats_output(
            self, stats_output: Dict[Tuple[str, str], pd.DataFrame]) -> bool:
        '''Write stats to file.'''
        try:
            with open(self._path / self._csv_filename, 'a+', newline='') as f:
                for s in stats_output.values():
                    print(s)
                    s.to_csv(f, mode='a+', header=False)
        except (PermissionError,):
            return False

        return True



# -*- coding: utf-8 -*-
'''
healthcare subjects module for reinforcement learning
=====================================================

This module provides different healthcare subjects in reinforcement learning
context.

Classes
-------
HealthSubject:
    A base `Subject` that accepts a `Patient`, administers the doses,
    and record the measurement of the outcome.

Warfarin:
    A `Subject` for warfarin that uses `PatientWarfarinRavvaz` and
    `healthcare.hamberg_pkpd`.
'''

from .health_subject import HealthSubject  # noqa: W0611
from .warfarin import Warfarin  # noqa: W0611



# -*- coding: utf-8 -*-
'''
learners module for reinforcement learning
==========================================

This module provides different learners in reinforcement learning context.

Classes
-------
Learner:
    The base class of all `learner` classes

Dense:
    A fully-connected neural net

QLookupTable:
    A simple lookup table for Q-learning

LearningRateScheduler:
    Base class for learning rate schedulers

ConstantLearningRate:
    A class that returns a constant learning rate
'''

from .learning_rate_schedulers import (  # noqa: W0611
    LearningRateScheduler, ConstantLearningRate)

from .learner import Learner  # noqa: W0611
from .lookup_table import QLookupTable, TableEntry  # noqa: W0611
from .fully_connected_neural_net import Dense  # noqa: W0611



# -*- coding: utf-8 -*-
'''
RandomAgent class
=================

An agent that randomly chooses an action
'''

import random
from typing import Any, Optional, Tuple

from reil.agents.no_learn_agent import NoLearnAgent
from reil.datatypes.feature import FeatureArray


class RandomAgent(NoLearnAgent):
    '''
    An agent that acts randomly.
    '''

    def __init__(self,
                 default_actions: Tuple[FeatureArray, ...] = (),
                 **kwargs: Any):
        super().__init__(default_actions=default_actions, **kwargs)

    def act(self,
            state: FeatureArray,
            subject_id: int,
            actions: Optional[Tuple[FeatureArray, ...]] = None,
            iteration: int = 0) -> FeatureArray:
        '''
        Return a random action.

        Arguments
        ---------
        state:
            The state for which the action should be returned.

        actions:
            The set of possible actions to choose from.

        iteration:
            The iteration in which the agent is acting.

        Returns
        -------
        :
            The action
        '''
        return random.choice(actions or self._default_actions)



# -*- coding: utf-8 -*-
'''
patient module for healthcare
=============================

This module provides classes to model a patient.

Classes
-------
Patient:
    The base class of all patient classes

PatientWarfarinRavvaz:
    A warfarin patient model with features and parameters of
    Ravvaz et al. 2016.
'''

from .patient import Patient  # noqa: W0611
from .patient_warfarin_ravvaz import PatientWarfarinRavvaz  # noqa: W0611

from . import agents, subjects  # noqa: W0611



import unittest
from typing import Any, Dict, Iterator, Union

import pandas as pd
from openpyxl import load_workbook
from reil.healthcare.dosing_protocols.warfarin import Aurora


class DummyPatients:
    def __init__(self, filename: str) -> None:
        wb = load_workbook(filename, data_only=True)
        self._trajectories = self.read_table(
            wb, 'trajectories', 'trajectories')
        self._trajectories['INR'] = self._trajectories.INR.apply(
            round, args=(2,))
        self._patient_info = self.read_table(
            wb, 'patient_info', 'patient_info', 'ID')

    @staticmethod
    def read_table(workbook, sheet_name, table_name, index=None):
        ref = workbook[sheet_name]._tables[table_name].ref
        content = [
            [cell.value for cell in ent]
            for ent in workbook[sheet_name][ref]
        ]
        header = content[0]
        rest = content[1:]

        df = pd.DataFrame(rest, columns=header)
        if index:
            df = df.set_index(index)

        return df

    def simulate(self, dose: float = -1, interval: int = -1) -> Iterator:
        for _id, info in self._patient_info.iterrows():
            patient: Dict[Union[str, int], Any] = dict(info.items())
            patient['ID'] = _id
            trajectory = self._trajectories[self._trajectories.ID == _id][
                ['day', 'INR', 'dose', 'interval']]
            # print(f'patient: {_id}')

            for i in trajectory.index:
                data = trajectory.loc[:i]  # type: ignore

                patient['day'] = data.day.iat[-1]
                patient['INR_history'] = list(data.INR)
                patient['dose_history'] = list(data.dose.iloc[:-1])
                patient['interval_history'] = list(data.interval.iloc[:-1])

                yield patient, data.dose.iat[-1], data.interval.iat[-1]


class testAurora(unittest.TestCase):
    def test_aurora(self) -> None:
        aurora = Aurora()

        patients = DummyPatients('./tests/data/aurora_sample_dosing.xlsx')

        additional_info = {}
        dose, interval = -1, -1
        _id = -1
        for p, d, i in patients.simulate(
                dose=dose, interval=interval):
            if p['ID'] != _id:
                _id = p['ID']
                additional_info = {}

            dosing_decision, additional_info = aurora.prescribe(
                patient=p, additional_info=additional_info)
            if additional_info['skip_dose']:
                dosing_decision, additional_info = aurora.prescribe(
                    patient=p, additional_info=additional_info)
            try:
                self.assertAlmostEqual(d, dosing_decision.dose)
                self.assertEqual(i, dosing_decision.duration)
            except AssertionError:
                print(d, dose, i, interval, '\n', p, '\n', additional_info)
                raise


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
'''
Feature class
=============

A datatype that accepts initial value and feature generator, and generates
new values.
'''
from __future__ import annotations

import dataclasses
from typing import Any, Callable, Optional, Tuple, Union

from reil.datatypes import reildata


@dataclasses.dataclass(frozen=True)
class Feature_old(reildata.FeatureGenerator):
    '''
    A datatype that accepts initial value and feature generator, and generates
    new values.

    Attributes
    ----------
    is_numerical:
        Is the feature numerical?

    value:
        The currect value of the feature.

    randomized:
        Whether the generator should produce random values.

    generator:
        A function that accepts feature characteristics and generates a new
        value

    lower:
        The lower bound for numerical features.

    upper:
        The upper bound for numerical features.

    mean:
        Mean of the distribution for numerical features.

    stdev:
        Standard Deviation of the distribution for numerical features.

    categories:
        A list of possible values for categorical features.

    probabilities:
        A list of probabilities corresponding to each possible value
        for categorical features.
    '''
    # value: Optional[Any] = None

    def __post_init__(self):
        super().__post_init__()
        if not self.is_numerical and self.probabilities is not None:
            if abs(sum(self.probabilities) - 1.0) > 1e-6:
                raise ValueError('probabilities should add up to 1.0.'
                                 f'Got {sum(self.probabilities)}')
            if self.categories is None:
                raise ValueError(
                    'probabilities cannot be set for None categories.')
            if len(self.probabilities) != len(self.categories):
                raise ValueError('Size mismatch. '
                                 f'{len(self.categories)} categories vs. '
                                 f'{len(self.categories)} probabilities')

    @classmethod
    def categorical(cls,
                    name,
                    # value: Optional[Any] = None,
                    categories: Optional[Tuple[Any, ...]] = None,
                    probabilities: Optional[Tuple[float, ...]] = None,
                    randomized: Optional[bool] = None,
                    generator: Optional[Callable[[Feature_old], Any]] = None):
        '''
        Create a categorical Feature.

        Arguments
        ---------
        value:
            The initial value of the feature.

        randomized:
            Whether the generator should produce random values.

        generator:
            A function that gets feature characteristics and generates a new
            value

        categories:
            A list of possible values.

        probabilities:
            A list of probabilities corresponding to each possible value.
        '''
        instance = cls(name=name,
                       is_numerical=False,
                       categories=categories,
                       probabilities=probabilities,
                       randomized=randomized,
                       generator=generator)

        return instance

    @classmethod
    def numerical(cls,
                  name,
                  lower: Optional[Union[int, float]] = None,
                  upper: Optional[Union[int, float]] = None,
                  mean: Optional[Union[int, float]] = None,
                  stdev: Optional[Union[int, float]] = None,
                  generator: Optional[Callable[[Feature_old], Any]] = None,
                  randomized: Optional[bool] = None):
        '''
        Create a numerical Feature.

        Arguments
        ---------
        value:
            The currect value of the feature.

        randomized:
            Whether the generator should produce random values.

        generator:
            A function that gets feature characteristics and generates a new
            value

        lower:
            The lower bound.

        upper:
            The upper bound.

        mean:
            Mean of the distribution.

        stdev:
            Standard Deviation of the distribution.
        '''
        instance = cls(name=name,
                       is_numerical=True,
                       lower=lower,
                       upper=upper,
                       mean=mean,
                       stdev=stdev,
                       generator=generator,
                       randomized=randomized)

        return instance

    def generate(self):
        '''
        Generate a new value using the generator.
        '''
        try:
            return self.generator(self)
        except TypeError:
            return self.generator



# -*- coding: utf-8 -*-
'''
QLearning class
===============

A Q-learning `agent`.


'''

from typing import Any, Optional, Tuple, Union

import numpy as np
from reil import stateful
from reil.agents.agent import Agent, TrainingData
from reil.datatypes.buffers import Buffer
from reil.datatypes.feature import FeatureArray
from reil.learners import Learner
from reil.utils.exploration_strategies import ExplorationStrategy

Feature_or_Tuple_of_Feature = Union[Tuple[FeatureArray, ...], FeatureArray]


class QLearning(Agent[float]):
    '''
    A Q-learning `agent`.
    '''

    def __init__(self,
                 learner: Learner[float],
                 buffer: Buffer[FeatureArray, float],
                 exploration_strategy: ExplorationStrategy,
                 method: str = 'backward',
                 **kwargs: Any):
        '''
        Arguments
        ---------
        learner:
            the `Learner` object that does the learning.

        exploration_strategy:
            an `ExplorationStrategy` object that determines
            whether the `action` should be exploratory or not for a given
            `state` at a given `iteration`.

        discount_factor:
            by what factor should future rewards be discounted?

        default_actions:
            a tuple of default actions.

        training_mode:
            whether the agent is in training mode or not.

        tie_breaker:
            how to choose the `action` if more than one is candidate
            to be chosen.
        '''
        super().__init__(learner=learner,
                         exploration_strategy=exploration_strategy,
                         **kwargs)

        self._method = method.lower()
        if self._method not in ('backward', 'forward'):
            self._logger.warning(
                f'method {method} is not acceptable. Should be '
                'either "forward" or "backward". Will use "backward".')
            self._method = 'backward'

        self._buffer = buffer
        self._buffer.setup(buffer_names=['X', 'Y'])

    @classmethod
    def _empty_instance(cls):  # type: ignore
        class MockBuffer:
            def setup(self, **kwargs: Any):
                pass

        return cls(None, MockBuffer(), None)  # type: ignore

    def _q(self,
           state: Feature_or_Tuple_of_Feature,
           action: Optional[Feature_or_Tuple_of_Feature] = None
           ) -> Tuple[float, ...]:
        '''
        Return the Q-value of `state` `action` pairs.

        Arguments
        ---------
        state:
            One state or a list of states for which Q-value is returned.

        action:
            One action or a list of actions for which Q-value is returned.
            If not supplied, `default_actions` will be used.

        Notes
        -----
        If one of state or action is one item, it will be broadcasted to
        match the size of the other one. If both are lists, the should match in
        size.


        :meta public:
        '''
        state_list = [state] if isinstance(state, FeatureArray) else state
        len_state = len(state_list)

        if action is None:
            action_list = self._default_actions
        else:
            action_list = ([action] if isinstance(action, FeatureArray)
                           else action)

        len_action = len(action_list)

        if len_state == len_action:
            x = tuple(state_list[i] + action_list[i]
                      for i in range(len_state))
        elif len_action == 1:
            x = tuple(state_list[i] + action_list[0]
                      for i in range(len_state))
        elif len_state == 1:
            x = tuple(state_list[0] + action_list[i]
                      for i in range(len_action))
        else:
            raise ValueError(
                'State and action should be of the same size'
                ' or at least one should be of size one.')

        return self._learner.predict(x)

    def _max_q(self, state: Feature_or_Tuple_of_Feature) -> float:
        '''
        Return `max(Q)` of one state or a list of states.

        Arguments
        ---------
        state:
            One state or a list of states for which MAX(Q) is returned.


        :meta public:
        '''
        try:
            q_values = self._q(state)
            max_q: float = np.max(q_values)
        except ValueError:
            max_q = 0.0

        return max_q

    def _prepare_training(
            self, history: stateful.History) -> TrainingData[float]:
        '''
        Use `history` to create the training set in the form of `X` and `y`
        vectors.

        Arguments
        ---------
        history:
            a `History` object from which the `agent` learns.

        Returns
        -------
        :
            a `TrainingData` object that contains `X` and 'y` vectors

        :meta public:
        '''
        state: FeatureArray
        action: FeatureArray
        next_state: FeatureArray
        reward: float

        discount_factor = self._discount_factor

        for h in history[:-1]:
            if h.state is None or h.action is None:
                raise ValueError(f'state and action cannot be None.\n{h}')

        # When history is one complete trajectory, the last observation
        # contains only the terminal state. In this case, we don't have an
        # action and a reward for the last observation, so we do not compute
        # its new Q value.
        if history[-1].action is None:
            active_history = history[:-1]
        else:
            active_history = history

        if self._method == 'forward':
            for i, h in enumerate(active_history):
                state: FeatureArray = h.state  # type: ignore
                action: FeatureArray = h.action  # type: ignore
                reward = h.reward or 0.0

                try:
                    next_state = history[i+1].state  # type: ignore
                    new_q = reward + discount_factor * \
                        self._max_q(next_state)
                except IndexError:
                    new_q = reward

                self._buffer.add(
                    {'X': state + action, 'Y': new_q})

        else:  # backward
            q_list = [0.0] * len(active_history)
            for i in range(len(active_history)-2, -1, -1):
                state = active_history[i].state  # type: ignore
                action = active_history[i].action  # type: ignore
                reward = active_history[i].reward or 0.0
                q_list[i] = reward + discount_factor * q_list[i+1]

                self._buffer.add(
                    {'X': state + action, 'Y': q_list[i]})

        temp = self._buffer.pick()

        return temp['X'], temp['Y']  # type: ignore

    def best_actions(self,
                     state: FeatureArray,
                     actions: Tuple[FeatureArray, ...]
                     ) -> Tuple[FeatureArray, ...]:
        '''
        Find the best `action`s for the given `state`.

        Arguments
        ---------
        state:
            The state for which the action should be returned.

        actions:
            The set of possible actions to choose from.

        Returns
        -------
        :
            A list of best actions.
        '''
        # None is used to avoid redundant normalization of default_actions
        q_values = self._q(state, None if actions ==
                           self._default_actions else actions)
        max_q = np.max(q_values)
        result = tuple(
            actions[i]  # type: ignore
            for i in np.nonzero(q_values == max_q)[0])

        return result

    def reset(self) -> None:
        '''Resets the agent at the end of a learning iteration.'''
        super().reset()
        self._buffer.reset()



# -*- coding: utf-8 -*-
'''
subjects module for reinforcement learning
==========================================

This module provides different subjects in reinforcement learning
context.

Classes
-------
Subject:
    The base class of all `subject` classes.

SubjectDemon:
    A class that allows a `subject` to be manipulated.

MNKGame:
    A simple game consisting of an m-by-n board.
    Each player should make a horizontal, vertical, or
    diagonal sequence of size k to win the game.

TicTacToe:
    Standard Tic-Tac-Toe game.

FrozenLake:
    (disabled)
    A frozen lake with cracks in it! (uses legacy ValueSet instead of
    FeatureArray)

WindyGridworld:
    (disabled)
    A grid with displacement of `agent` (as if wind blows)
    (uses legacy ValueSet instead of FeatureArray)
'''

from .subject import Subject  # noqa: W0611
from .subject_demon import SubjectDemon, Modifier  # noqa: W0611
from .mnkgame import MNKGame  # noqa: W0611
from .tic_tac_toe import TicTacToe  # noqa: W0611
# from .windy_gridworld import WindyGridworld  # noqa: W0611
# from .frozen_lake import FrozenLake  # noqa: W0611



# -*- coding: utf-8 -*-
'''
learning_rate_schedulers module
===============================

A module that provides different learning rate schedulers.

classes
-------
LearningRateScheduler:
    The base class of all schedulers. Accepts a bi-variate function.

ConstantLearningRate:
    A learning rate scheduler with constant rate.
'''
from typing import Callable
from reil import reilbase


class LearningRateScheduler(reilbase.ReilBase):
    '''
    A class that gets an initial learning rate and a function to determine the
    rate based on iteration and previous rate.
    '''
    def __init__(self, initial_lr: float,
                 new_rate_function: Callable[[int, float], float]) -> None:
        '''
        Arguments
        ---------
        initial_lr:
            The initial learning rate.

        new_rate_function:
            A function that accepts iteration and current learning
            rate and returns a new learning rate.
        '''
        self.initial_lr = initial_lr
        self._lambda_func = new_rate_function

    @classmethod
    def _empty_instance(cls):
        return cls(0.0, lambda e, lr: lr)

    def new_rate(self, iteration: int, current_lr: float) -> float:
        '''
        Determine the new rate based on `iteration` and current learning rate.

        Arguments
        ---------
        iteration:
            The current training iteration.

        current_lr:
            The current learning rate.

        Returns
        -------
        :
            The new learning rate.

        '''
        return self._lambda_func(iteration, current_lr)


class ConstantLearningRate(LearningRateScheduler):
    '''
    A `LearningRateScheduler` with constant rate.
    '''
    def __init__(self, initial_lr: float) -> None:
        '''
        Arguments
        ---------
        initial_lr:
            The initial learning rate.
        '''
        super().__init__(initial_lr, lambda e, lr: initial_lr)



import unittest
from random import randint, sample

from reil.datatypes import feature


class testFeature(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls._numerical_data = [randint(-100, 100) for _ in range(10)]
        cls._categorical_data = sample(list('ABCDEFGHIJKLMNOPQRSTUVWXYZ'), 10)

    def test_create_feature(self):
        rl_value = feature.Feature(
            name='test', value=self._numerical_data, is_numerical=True)
        self.assertEqual(rl_value.name, 'test')
        self.assertEqual(rl_value.value, self._numerical_data)
        self.assertEqual(rl_value.is_numerical, True)

        lower = min(self._numerical_data)
        upper = max(self._numerical_data)

        def midpoint(f):
            return (f.lower + f.upper)/2

        numerical = feature.FeatureGenerator.numerical(
            name='num', lower=lower,
            upper=upper, generator=midpoint, randomized=True)

        self.assertEqual(numerical.name, 'num')
        self.assertEqual(numerical().value, (lower + upper)/2)
        self.assertEqual(numerical.is_numerical, True)

        categorical = feature.FeatureGenerator.categorical(
            name='cat', categories=('A', 'B', 'C'), probabilities=(.3, .6, .1),
            generator=lambda f: 'A', randomized=True)

        self.assertEqual(categorical.name, 'cat')
        self.assertEqual(categorical().value, 'A')
        self.assertEqual(categorical.is_numerical, False)

    def test_add_features(self):
        rl_values = [feature.Feature(
            name='test', value=v, is_numerical=True)
            for v in self._numerical_data]
        rl_new = sum(rl_values[1:], rl_values[0])
        self.assertEqual(rl_new.name, 'test')
        self.assertEqual(rl_new.value, sum(self._numerical_data))
        self.assertEqual(rl_new.is_numerical, True)

    def test_add_numerical(self):
        numericals = feature.FeatureGenerator.numerical(name='test')
        rl_values = [numericals(v) for v in self._numerical_data]

        rl_new = sum(rl_values[1:], rl_values[0])
        self.assertEqual(rl_new.name, 'test')
        self.assertEqual(rl_new.value, sum(self._numerical_data))
        self.assertEqual(rl_new.is_numerical, True)

        rl_values = [numericals((v,)) for v in self._numerical_data]

        rl_new = sum(rl_values[1:], rl_values[0])
        self.assertEqual(rl_new.name, 'test')
        self.assertEqual(rl_new.value, tuple(self._numerical_data))
        self.assertEqual(rl_new.is_numerical, True)

    def test_add_categorical(self):
        categoricals = feature.FeatureGenerator.categorical(
            name='test', categories=tuple(self._categorical_data))
        rl_values = [categoricals(v) for v in self._categorical_data]
        self.assertEqual(
            rl_values[0].normalized,
            tuple([1] + [0]*(len(self._categorical_data) - 2)))
        self.assertEqual(
            rl_values[-1].normalized,
            tuple([0]*(len(self._categorical_data) - 1)))

        rl_new = sum(rl_values[1:], rl_values[0])
        self.assertEqual(rl_new.name, 'test')
        self.assertEqual(rl_new.value, ''.join(self._categorical_data))
        self.assertEqual(rl_new.is_numerical, False)

        rl_values = [categoricals((v,)) for v in self._categorical_data]

        rl_new = sum(rl_values[1:], rl_values[0])
        self.assertEqual(rl_new.name, 'test')
        self.assertEqual(rl_new.value, tuple(self._categorical_data))
        self.assertEqual(rl_new.is_numerical, False)

    def test_allow_missing(self):
        categoricals = feature.FeatureGenerator.categorical(
            name='test', categories=tuple(self._categorical_data),
            allow_missing=True)
        rl_values = [categoricals(v) for v in self._categorical_data]
        self.assertEqual(
            rl_values[0].normalized,
            tuple([1] + [0]*(len(self._categorical_data) - 1)))
        self.assertEqual(
            rl_values[-1].normalized,
            tuple([0]*(len(self._categorical_data) - 1) + [1]))
        self.assertEqual(
            categoricals(feature.MISSING).normalized,
            tuple([0]*(len(self._categorical_data))))

    def test_feature_array(self):
        categoricals = feature.FeatureGenerator.categorical(
            name='cats', categories=tuple(self._categorical_data))
        numericals = feature.FeatureGenerator.numerical(
            name='nums', lower=min(self._numerical_data),
            upper=max(self._numerical_data))

        with self.assertRaises(KeyError):
            feature.FeatureArray(
                [categoricals(v) for v in self._categorical_data])
            # print(rl_array.categories)

        reilarray = feature.FeatureArray([
            categoricals(self._categorical_data[0]),
            numericals(self._numerical_data[0]),
        ])

        self.assertEqual(reilarray.lower,
                         {reilarray['cats'].name: reilarray['cats'].lower,
                          reilarray['nums'].name: reilarray['nums'].lower})
        self.assertEqual(reilarray.upper,
                         {reilarray['cats'].name: reilarray['cats'].upper,
                          reilarray['nums'].name: reilarray['nums'].upper})
        self.assertEqual(
            reilarray.categories,
            {reilarray['cats'].name: reilarray['cats'].categories,
             reilarray['nums'].name: reilarray['nums'].categories})

        self.assertEqual(reilarray.normalized.flatten(),
                         list(reilarray['cats'].normalized) +
                         [reilarray['nums'].normalized])

        self.assertEqual(
            reilarray,
            feature.FeatureArray([categoricals(self._categorical_data[0])]) +
            feature.FeatureArray([numericals(self._numerical_data[0])])
            )

    def test_change_to_missing(self):
        categoricals = feature.FeatureGenerator.categorical(
            name='test', categories=tuple(self._categorical_data),
            allow_missing=True)
        self.assertEqual(
            feature.change_to_missing(categoricals(
                self._categorical_data[0])).normalized,
            tuple([0]*len(self._categorical_data)))

        array = feature.FeatureArray(
            feature.FeatureGenerator.categorical(
                name=v, categories=tuple(self._categorical_data),
                allow_missing=True)(v)
            for v in self._categorical_data)
        missing_array = feature.change_array_to_missing(array)
        self.assertEqual(
            next(iter(missing_array)).normalized,
            tuple([0]*len(self._categorical_data)))

        with self.assertRaises(TypeError):
            feature.change_to_missing(
                feature.FeatureGenerator.categorical(
                    name='test', categories=tuple(self._categorical_data),
                    allow_missing=False)(self._categorical_data[0]))

        with self.assertRaises(ValueError):
            feature.change_to_missing(
                feature.FeatureGenerator.categorical(
                    name='test',
                    allow_missing=True)(self._categorical_data[0]))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
'''
WindyGridworld class
==============

This class creates a gridworld (board) in which one square is the goal.
The agent starts from a location and should find the fastest route to the goal.
Wind is a fixed amount of padding based on the location of the agent on the board.


'''


from random import choice

from ..utils import MNKBoard
from ..subjects import Subject
from ..legacy import ValueSet


def main():
    board = WindyGridworld(h_wind=[1, 1, 1, 1, 1], move_pattern='Q')
    _ = board.register('P1')
    for _ in range(10):
        print(board.state)
        my_action = choice(board.possible_actions)
        board.take_effect(my_action, 1)
        print(my_action.value)
        print(f'{board}')


class WindyGridworld(MNKBoard, Subject):
    '''
    Build an m-by-n grid (using mnkboard super class) in which 1 player can play.
    Player wins if it can get to the goal square.

    Attributes
    ----------
        is_terminated: whether the game finished or not.
        possible_actions: a list of possible actions.

    Methods
    -------
        register: register the player and return its ID or return ID of current player.
        take_effect: moves the player on the grid.
        reset: clears the grid.
    '''
    def __init__(self, **kwargs):
        '''
        Initialize an instance of windy gridworld.

        Arguments
        ---------
            dim: dimensions of the grid as a tuple. (default=(5, 5))
            start: (row, column) of the starting square. (default=(0, 0))
            goal: (row, column) of the goal square. (default=(4, 4))
            h_wind: a list of size row that determines the strenghs of the wind. (default=0)
            v_wind: a list of size column that determines the strenghs of the wind. (default=0)
            move_pattern: whether to choose 'R'ook moves or 'Q'ueen moves. (default='R')
            state_type: board (zero one list), tuple ()
        '''
        self._dim, self._start, self._goal = (5, 5), (0, 0), (4, 4)
        self._move_pattern = 'R'
        self._h_wind, self._v_wind = [0]*5, [0]*5
        self.set_params(**kwargs)
        super().__init__(**kwargs)

        if self._move_pattern.upper() == 'Q':
            moves = ['U', 'D', 'R', 'L', 'UR', 'UL', 'DR', 'DL']
        elif self._move_pattern.upper() == 'R':
            moves = ['U', 'D', 'R', 'L']

        self._default_moves = ValueSet(moves, binary=lambda x:(moves.index(x), len(moves))).as_valueset_array()

        super().__init__(m=self._dim[0], n=self._dim[1], players=1)
        self.reset()

    @property
    def is_terminated(self):
        '''Return True if the player get to the goal.'''
        return [*self._player_location] == [*self._goal]

    @property
    def state(self):
        '''Return the state of the board as a ValueSet.'''
        return ValueSet(self._player_location, binary=lambda x: (x[0]*self._dim[1] + x[1], len(self._board)))

    @property
    def possible_actions(self):
        '''Return the set of possible moves.'''
        return self._default_moves

    def register(self, player_name):
        '''
        Register an agent and return its ID.
        
        If the agent is new, a new ID is generated and the agent_name is added to agent_list.
        Arguments
        ---------
            agent_name: the name of the agent to be registered.

        Raises ValueError for an attempt to register more than one agent.
        '''
        if len(self._agent_list) == 0:
            return Subject.register(self, player_name)
        raise ValueError('Windy Gridworld only accepts one player.')

    def take_effect(self, action, _id=None):
        '''
        Move according to the action.

        Arguments
        ---------
            _id: ID of the player.
            action: the location in which the piece is set. Can be either in index format or row column format.
        '''
        row, column = [*self._player_location]
        max_row, max_column = self._dim[0]-1, self._dim[1]-1
        MNKBoard.clear_square(self, row=row, column=column)

        a = str(*action.value)
        if a in ['U', 'UR', 'UL']:
            self._player_location[0] = row - 1  # min(max(row-1+self._v_wind[column], 0), max_row)
        if a in ['D', 'DR', 'DL']:
            self._player_location[0] = row + 1  # max(min(row+1+self._v_wind[column], max_row), 0)
        if a in ['L', 'UL', 'DL']:
            self._player_location[1] = column - 1  # min(max(column-1+self._h_wind[row], 0), max_column)
        if a in ['R', 'UR', 'DR']:
            self._player_location[1] = column + 1  # max(min(column+1+self._h_wind[row], max_column), 0)

        self._player_location[0] = min(max(self._player_location[0]+self._v_wind[column], 0), max_row)
        self._player_location[1] = min(max(self._player_location[1]+self._h_wind[row], 0), max_column)

        MNKBoard.set_piece(self, player=1, row=self._player_location[0], column=self._player_location[1])
        if self.is_terminated:
            return 1
        return -1

    def reset(self):
        '''Clear the board and update board_status.'''
        MNKBoard.reset(self)
        self._player_location = [*self._start]
        MNKBoard.set_piece(self, player=1, row=self._player_location[0], column=self._player_location[1])

    def __repr__(self):
        return 'WindyGridworld'

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
'''
healthcare mathematical models module
=====================================

This module provides mathematical models, e.g. PK/PDs, cancer PDEs, etc.

Classes
-------
HealthMathModel:
    The base class for healthcare math models.

HambergPKPD:
    Hamberg's warfarin PK/PD model.
'''

from .health_math_model import HealthMathModel  # noqa: W0611
from .hamberg_pkpd import HambergPKPD  # noqa: W0611
